Type: Package
Package: tidynorm
Title: Tools for Tidy Vowel Normalization
Version: 0.4.1
Authors@R: c(
    person("Josef", "Fruehwald", , "JoFrhwld@gmail.com", role = c("cre", "aut", "cph"))
  )
Description: An implementation of tidy speaker vowel normalization.
    This includes generic functions for defining new normalization methods for
    points, formant tracks, and Discrete Cosine Transform coefficients, as well
    as convenience functions implementing established normalization methods.
    References for the implemented methods are:
    Johnson, Keith (2020) <doi:10.5334/labphon.196>
    Lobanov, Boris (1971) <doi:10.1121/1.1912396>
    Nearey, Terrance M. (1978) <https://sites.ualberta.ca/~tnearey/Nearey1978_compressed.pdf>
    Syrdal, Ann K., and Gopal, H. S. (1986) <doi:10.1121/1.393381>
    Watt, Dominic, and Fabricius, Anne (2002) <https://www.latl.leeds.ac.uk/article/evaluation-of-a-technique-for-improving-the-mapping-of-multiple-speakers-vowel-spaces-in-the-f1-f2-plane/>.
License: GPL (>= 3)
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.2
Imports: checkmate, cli, dplyr, glue, options, purrr, Rcpp, rlang,
        stringr, tidyr, tidyselect
URL: https://jofrhwld.github.io/tidynorm/,
        https://github.com/JoFrhwld/tidynorm
Depends: R (>= 4.1)
Suggests: ggdensity, ggplot2, knitr, magick, quarto, reticulate,
        rmarkdown, testthat (>= 3.0.0), tibble
Config/testthat/edition: 3
Config/testthat/parallel: true
VignetteBuilder: quarto
LinkingTo: Rcpp, RcppArmadillo
BugReports: https://github.com/JoFrhwld/tidynorm/issues
NeedsCompilation: yes
Packaged: 2026-05-10 15:10:56 UTC; joseffruehwald
Author: Josef Fruehwald [cre, aut, cph]
Maintainer: Josef Fruehwald <JoFrhwld@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-10 15:30:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-10 15:48:15 UTC; unix
Archs: tidynorm.so.dSYM

Type: Package
Package: tidynorm
Title: Tools for Tidy Vowel Normalization
Version: 0.4.1
Authors@R: c(
    person("Josef", "Fruehwald", , "JoFrhwld@gmail.com", role = c("cre", "aut", "cph"))
  )
Description: An implementation of tidy speaker vowel normalization.
    This includes generic functions for defining new normalization methods for
    points, formant tracks, and Discrete Cosine Transform coefficients, as well
    as convenience functions implementing established normalization methods.
    References for the implemented methods are:
    Johnson, Keith (2020) <doi:10.5334/labphon.196>
    Lobanov, Boris (1971) <doi:10.1121/1.1912396>
    Nearey, Terrance M. (1978) <https://sites.ualberta.ca/~tnearey/Nearey1978_compressed.pdf>
    Syrdal, Ann K., and Gopal, H. S. (1986) <doi:10.1121/1.393381>
    Watt, Dominic, and Fabricius, Anne (2002) <https://www.latl.leeds.ac.uk/article/evaluation-of-a-technique-for-improving-the-mapping-of-multiple-speakers-vowel-spaces-in-the-f1-f2-plane/>.
License: GPL (>= 3)
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.2
Imports: checkmate, cli, dplyr, glue, options, purrr, Rcpp, rlang,
        stringr, tidyr, tidyselect
URL: https://jofrhwld.github.io/tidynorm/,
        https://github.com/JoFrhwld/tidynorm
Depends: R (>= 4.1)
Suggests: ggdensity, ggplot2, knitr, magick, quarto, reticulate,
        rmarkdown, testthat (>= 3.0.0), tibble
Config/testthat/edition: 3
Config/testthat/parallel: true
VignetteBuilder: quarto
LinkingTo: Rcpp, RcppArmadillo
BugReports: https://github.com/JoFrhwld/tidynorm/issues
NeedsCompilation: yes
Packaged: 2026-05-10 15:10:56 UTC; joseffruehwald
Author: Josef Fruehwald [cre, aut, cph]
Maintainer: Josef Fruehwald <JoFrhwld@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-10 15:30:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-10 17:12:37 UTC; unix
Archs: tidynorm.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(dct,matrix)
S3method(dct,numeric)
S3method(idct,matrix)
S3method(idct,numeric)
export(bark_to_hz)
export(check_norm)
export(dct)
export(dct_basis)
export(hz_to_bark)
export(hz_to_mel)
export(idct)
export(idct_accel)
export(idct_rate)
export(mel_to_hz)
export(norm_barkz)
export(norm_dct_barkz)
export(norm_dct_deltaF)
export(norm_dct_generic)
export(norm_dct_lobanov)
export(norm_dct_nearey)
export(norm_dct_wattfab)
export(norm_deltaF)
export(norm_generic)
export(norm_lobanov)
export(norm_nearey)
export(norm_track_barkz)
export(norm_track_deltaF)
export(norm_track_generic)
export(norm_track_lobanov)
export(norm_track_nearey)
export(norm_track_wattfab)
export(norm_wattfab)
export(reframe_with_dct)
export(reframe_with_dct_smooth)
export(reframe_with_idct)
export(tidynorm_options)
import(cli)
import(dplyr)
import(rlang)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(checkmate,check_character)
importFrom(options,opt)
importFrom(rlang,`!!`)
importFrom(stats,sd)
useDynLib(tidynorm, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# tidynorm <a href="https://jofrhwld.github.io/tidynorm/"><img src="man/figures/logo.png" alt="tidynorm website" align="right" height="138"/></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/JoFrhwld/tidynorm/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/JoFrhwld/tidynorm/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/tidynorm)](https://CRAN.R-project.org/package=tidynorm)

<!-- badges: end -->

The goal of `{tidynorm}` is to provide convenient and tidy functions to
normalize vowel formant data.

## Installation

You can install tidynorm like so

``` r
install.packages("tidynorm")
```

You can install the development version of tidynorm like so:

``` r
## if you need to install `remotes`
# install.packages("remotes")
remotes::install_github("jofrhwld/tidynorm")
```

## Example

Vowel formant frequencies are heavily influenced by vocal tract length
differences between speakers. Equivalent vowels between speakers can
have dramatically different frequency locations.

``` r
library(tidynorm)
library(ggplot2)
```

<details class="code-fold">
<summary>Plotting Options</summary>

``` r
options(
  ggplot2.discrete.colour = c(
    lapply(
      1:6,
      \(x) c(
        "#4477AA", "#EE6677", "#228833",
        "#CCBB44", "#66CCEE", "#AA3377"
      )[1:x]
    )
  ),
  ggplot2.discrete.fill = c(
    lapply(
      1:6,
      \(x) c(
        "#4477AA", "#EE6677", "#228833",
        "#CCBB44", "#66CCEE", "#AA3377"
      )[1:x]
    )
  )
)

theme_set(
  theme_minimal(
    base_size = 16
  )
)
```

</details>

<details class="code-fold">
<summary>Plotting Code</summary>

``` r
ggplot(
  speaker_data,
  aes(
    F2, F1,
    color = speaker
  )
) +
  ggdensity::stat_hdr(
    probs = c(0.95, 0.8, 0.5),
    alpha = 1,
    fill = NA,
    linewidth = 1
  ) +
  scale_x_reverse() +
  scale_y_reverse() +
  coord_fixed() +
  labs(
    title = "unnormalized"
  )
```

</details>

<img src="man/figures/README-unnorm-1.png" style="width:80.0%"
data-fig-align="center"
data-fig-alt="A two dimensional density plot of speakers s01 and s02&#39;s vowel formant data. They are both roughly triangular, but s02&#39;s vowel space is smaller, and shifted up and back relative to  s01&#39;s. " />

The goal of `{tidynorm}` is to provide tidyverse-friendly and familiar
functions that will allow you to quickly normalize vowel formant data.
There are a number of built in functions based on conventional
normalization methods.

``` r
speaker_data |>
  norm_nearey(
    F1:F3,
    .by = speaker,
    .names = "{.formant}_nearey"
  ) ->
speaker_normalized
```

    #> Normalization info
    #> • normalized with `tidynorm::norm_nearey()`
    #> • normalized `F1`, `F2`, and `F3`
    #> • normalized values in `F1_nearey`, `F2_nearey`, and `F3_nearey`
    #> • grouped by `speaker`
    #> • within formant: FALSE
    #> • (.formant - mean(.formant, na.rm = T))/(1)

<details class="code-fold">
<summary>Plotting Code</summary>

``` r
speaker_normalized |>
  ggplot(
    aes(
      F2_nearey, F1_nearey,
      color = speaker
    )
  ) +
  ggdensity::stat_hdr(
    probs = c(0.95, 0.8, 0.5),
    alpha = 1,
    fill = NA,
    linewidth = 1
  ) +
  scale_x_reverse() +
  scale_y_reverse() +
  coord_fixed() +
  labs(
    title = "Nearey Normalized"
  )
```

</details>

<img src="man/figures/README-norm-1.png" style="width:60.0%"
data-fig-align="center"
data-fig-alt="A two dimensional density plot of speakers s01 and s02&#39;s vowel formant data. They are both roughly triangular, and highly overlapping. " />

There is also a `tidynorm::norm_generic()` function to allow you to
define your own bespoke normalization methods. For example, a “robust
Nearey” normalization method using the median, instead of the mean,
could be done like so.

``` r
speaker_rnearey <- speaker_data |>
  norm_generic(
    F1:F3,
    .by = speaker,
    .by_formant = FALSE,
    .pre_trans = log,
    .L = median(.formant, na.rm = T),
    .names = "{.formant}_rnearey"
  )
```

    #> Normalization info
    #> • normalized with `tidynorm::norm_generic()`
    #> • normalized `F1`, `F2`, and `F3`
    #> • normalized values in `F1_rnearey`, `F2_rnearey`, and `F3_rnearey`
    #> • grouped by `speaker`
    #> • within formant: FALSE
    #> • (.formant - median(.formant, na.rm = T))/(1)

<details class="code-fold">
<summary>Plotting Code</summary>

``` r
speaker_rnearey |>
  ggplot(
    aes(
      F2_rnearey, F1_rnearey,
      color = speaker
    )
  ) +
  ggdensity::stat_hdr(
    probs = c(0.95, 0.8, 0.5),
    alpha = 1,
    fill = NA,
    linewidth = 1
  ) +
  scale_x_reverse() +
  scale_y_reverse() +
  coord_fixed() +
  labs(
    title = "Robust Nearey Normalized"
  )
```

</details>

<img src="man/figures/README-rnorm-1.png" style="width:60.0%"
data-fig-align="center"
data-fig-alt="A two dimensional density plot of speakers s01 and s02&#39;s vowel formant data. They are both roughly triangular, and highly overlapping. " />

library(tidynorm)
ggplot2_inst <- require(ggplot2)

speaker_data_wattfab <- speaker_data |>
  norm_wattfab(
    F1:F3,
    .by = speaker,
    .names = "{.formant}_wf"
  )

if (ggplot2_inst) {
  ggplot(
    speaker_data_wattfab,
    aes(
      F2_wf,
      F1_wf,
      color = speaker
    )
  ) +
    stat_density_2d(
      bins = 4
    ) +
    scale_color_brewer(
      palette = "Dark2"
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    coord_fixed()
}

library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

track_subset <- speaker_tracks |>
  filter(
    .by = c(speaker, id),
    if_all(
      F1:F3,
      .fns = \(x) mean(is.finite(x)) > 0.9
    ),
    row_number() %% 2 == 1
  )

track_norm <- track_subset |>
  norm_track_barkz(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .drop_orig = TRUE
  )

if (ggplot2_inst) {
  track_norm |>
    ggplot(
      aes(F2_bz, F1_bz, color = speaker)
    ) +
    stat_density_2d(bins = 4) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}


# returning the DCT coefficients
track_norm_dct <- track_subset |>
  norm_track_barkz(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .drop_orig = TRUE,
    .return_dct = TRUE,
    .names = "{.formant}_bz"
  )

track_norm_means <- track_norm_dct |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      ends_with("_bz"),
      mean
    )
  ) |>
  reframe_with_idct(
    ends_with("_bz"),
    .by = speaker,
    .token_id_col = vowel,
    .param_col = .param
  )


if (ggplot2_inst) {
  track_norm_means |>
    ggplot(
      aes(F2_bz, F1_bz, color = speaker)
    ) +
    geom_path(
      aes(
        group = interaction(speaker, vowel)
      )
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}

library(tidynorm)
library(dplyr)

ggplot2_inst <- require(ggplot2)

speaker_small <- filter(
  speaker_tracks,
  id == 0
)

speaker_dct_smooth <- speaker_small |>
  reframe_with_dct_smooth(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .order = 5
  )

if (ggplot2_inst) {
  speaker_small |>
    ggplot(
      aes(
        t, F1
      )
    ) +
    geom_point() +
    facet_wrap(
      ~speaker,
      scales = "free_x",
      ncol = 1
    ) +
    labs(
      title = "Original Data"
    )
}

if (ggplot2_inst) {
  speaker_dct_smooth |>
    ggplot(
      aes(
        t, F1
      )
    ) +
    geom_point() +
    facet_wrap(
      ~speaker,
      scales = "free_x",
      ncol = 1
    ) +
    labs(
      title = "Smoothed Data"
    )
}

library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

track_subset <- speaker_tracks |>
  filter(
    .by = c(speaker, id),
    if_all(
      F1:F3,
      .fns = \(x) mean(is.finite(x)) > 0.9
    ),
    row_number() %% 2 == 1
  )

track_dcts <- track_subset |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .order = 3
  )

track_norm <- track_dcts |>
  norm_dct_generic(
    F1:F3,
    .token_id_col = id,
    .by = speaker,
    .by_formant = TRUE,
    .L = median(.formant, na.rm = TRUE),
    .S = mad(.formant, na.rm = TRUE),
    .param_col = .param,
    .drop_orig = TRUE,
    .names = "{.formant}_mad"
  )

head(track_norm)

full_tracks <- track_norm |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      F1_mad:F3_mad,
      mean
    )
  ) |>
  reframe_with_idct(
    F1_mad:F3_mad,
    .by = c(speaker, vowel),
    .param_col = .param
  )

head(full_tracks)

if (ggplot2_inst) {
  ggplot(
    full_tracks,
    aes(F2_mad, F1_mad, color = speaker)
  ) +
    geom_path(
      aes(group = interaction(speaker, vowel))
    ) +
    scale_y_reverse() +
    scale_x_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}

library(tidynorm)
library(dplyr)

speaker_small <- filter(
  speaker_tracks,
  id == 0
)

speaker_dct <- reframe_with_dct(
  speaker_small,
  F1:F3,
  .by = speaker,
  .token_id_col = id,
  .time_col = t
)

head(
  speaker_dct
)

library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

speaker_dct <- speaker_tracks |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )

# Normalize DCT coefficients
speaker_dct_norm <- speaker_dct |>
  norm_dct_deltaF(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param
  )

# Apply average and apply inverse dct
# to plot tracks
track_norm_means <- speaker_dct_norm |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      ends_with("_df"),
      mean
    )
  ) |>
  reframe_with_idct(
    ends_with("_df"),
    .by = speaker,
    .token_id_col = vowel,
    .param_col = .param
  )


if (ggplot2_inst) {
  track_norm_means |>
    ggplot(
      aes(F2_df, F1_df, color = speaker)
    ) +
    geom_path(
      aes(
        group = interaction(speaker, vowel)
      )
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}

library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

speaker_small <- filter(
  speaker_tracks,
  id == 0
)

speaker_dct <- speaker_small |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .order = 5
  )

speaker_idct <- speaker_dct |>
  reframe_with_idct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param,
    .n = 20
  )

if (ggplot2_inst) {
  speaker_small |>
    mutate(
      .by = c(speaker, id),
      time_index = row_number()
    ) |>
    ggplot(
      aes(
        time_index, F1
      )
    ) +
    geom_point() +
    labs(
      title = "Original Data"
    )
}

if (ggplot2_inst) {
  speaker_idct |>
    ggplot(
      aes(
        .time, F1
      )
    ) +
    geom_point() +
    labs(
      title = "DCT Smooth Data"
    )
}

library(tidynorm)
library(dplyr)

speaker_data |>
  norm_generic(
    F1:F3,
    .by = speaker,
    .by_formant = TRUE,
    .L = median(.formant, na.rm = TRUE),
    .S = mad(.formant, na.rm = TRUE),
    .drop_orig = TRUE,
    .names = "{.formant}_mad"
  )

library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

speaker_dct <- speaker_tracks |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )

# Normalize DCT coefficients
speaker_dct_norm <- speaker_dct |>
  norm_dct_wattfab(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param
  )

# Apply average and apply inverse dct
# to plot tracks
track_norm_means <- speaker_dct_norm |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      ends_with("_wf"),
      mean
    )
  ) |>
  reframe_with_idct(
    ends_with("_wf"),
    .by = speaker,
    .token_id_col = vowel,
    .param_col = .param
  )


if (ggplot2_inst) {
  track_norm_means |>
    ggplot(
      aes(F2_wf, F1_wf, color = speaker)
    ) +
    geom_path(
      aes(
        group = interaction(speaker, vowel)
      )
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}

library(tidynorm)
ggplot2_inst <- require(ggplot2)

speaker_data_nearey <- speaker_data |>
  norm_nearey(
    F1:F3,
    .by = speaker,
    .by_formant = FALSE,
    .names = "{.formant}_nearey"
  )

if (ggplot2_inst) {
  ggplot(
    speaker_data_nearey,
    aes(
      F2_nearey,
      F1_nearey,
      color = speaker
    )
  ) +
    stat_density_2d(
      bins = 4
    ) +
    scale_color_brewer(
      palette = "Dark2"
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    coord_fixed() +
    labs(
      title = "Formant extrinsic"
    )
}

speaker_data_nearey2 <- speaker_data |>
  norm_nearey(
    F1:F3,
    .by = speaker,
    .by_formant = TRUE,
    .names = "{.formant}_nearey"
  )

if (ggplot2_inst) {
  ggplot(
    speaker_data_nearey2,
    aes(
      F2_nearey,
      F1_nearey,
      color = speaker
    )
  ) +
    stat_density_2d(
      bins = 4
    ) +
    scale_color_brewer(
      palette = "Dark2"
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    coord_fixed() +
    labs(
      title = "Formant intrinsic"
    )
}

#' Generic Normalization Procedure
#'
#' This is a generic normalization procedure with which
#' you can create your own normalization method.
#'
#' @param .data A data frame containing vowel formant data
#' @param ... [`<tidy-select>`][dplyr::dplyr_tidy_select] One or more unquoted
#' expressions separated by commas. These should target the vowel formant
#' data columns.
#' @param .by  [`<tidy-select>`][dplyr::dplyr_tidy_select] A selection of
#' columns to group by. Typically a column of speaker IDs.
#' @param .by_formant Whether or not the normalization method is formant
#' intrinsic.
#' @param .by_token Whether or not the normalization method is vowel intrinsic
#' @param .L An expression defining the location parameter.
#' See Details for more information.
#' @param .S An expression defining the scale parameter.
#' See Details for more information.
#' @param .pre_trans A function to apply to formant values before normalization.
#' @param .post_trans A function to apply to formant values after normalization.
#' @param .drop_orig Whether or not to drop the original formant data columns.
#' @param .keep_params Whether or not to keep the Location (`*_.L`) and Scale
#' (`*_.S`) normalization parameters
#' @param .names A [glue::glue()] expression for naming the normalized
#' data columns. The `"{.formant}"` portion corresponds to the name of the original
#' formant columns.
#' @param .call Used for internal purposes.
#'
#' @eval options::as_params(".silent" = "tidynorm.silent")
#'
#'
#' @details
#' The following `norm_*` procedures are built on top of `norm_generic()`.
#'
#' - [norm_lobanov]
#' - [norm_nearey]
#' - [norm_deltaF]
#' - [norm_wattfab]
#' - [norm_barkz]
#'
#' ### Location and Scale expressions
#' All normalization procedures built on [norm_generic] produce normalized
#' formant values (\eqn{\hat{F}}) by subtracting a location parameter
#' (\eqn{L}) and dividing by a scale parameter (\eqn{S}).
#'
#' \deqn{
#' \hat{F} = \frac{F-L}{S}
#' }
#'
#' The expressions for calculating \eqn{L} and \eqn{S} can be
#' passed to `.L` and `.S`, respectively. Available values for
#' these expressions are
#'
#' \describe{
#'  \item{`.formant`}{The original formant value}
#'  \item{`.formant_num`}{The number of the formant. (e.g. 1 for F1, 2 for F2 etc)}
#' }
#'
#' Along with any data columns from your original data.
#'
#' ### Pre and Post normalization transforms
#' To apply any transformations before or after normalization,
#' you can pass a function to `.pre_trans` and `.post_trans`.
#'
#' ### Formant In/Extrinsic Normalization
#' If `.by_formant` is `TRUE`, normalization will be formant intrinsic.
#' If `.by_formant` is `FALSE`, normalization will be formant extrinsic.
#'
#' ### Token In/Extrinsic Normalization
#' If `.by_token` is `TRUE`, normalization will be token intrinsic.
#' If `.by_token` is `FALSE`, normalization will be token extrinsic.
#'
#' @returns
#' A data frame of normalized formant values
#'
#' @example inst/examples/ex-norm_generic.R
#' @importFrom options opt
#'
#' @export
norm_generic <- function(
    .data,
    ...,
    .by = NULL,
    .by_formant = FALSE,
    .by_token = FALSE,
    .L = 0,
    .S = 1,
    .pre_trans = identity,
    .post_trans = identity,
    .drop_orig = FALSE,
    .keep_params = FALSE,
    .names = "{.formant}_n",
    .silent = opt("tidynorm.silent"),
    .call = caller_env()) {
  if (env_name(.call) == "global") {
    .call <- current_env()
  }

  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls, .call)

  targets <- expr(c(...))
  cols <- enquos(
    .by = .by
  )

  .names2 <- glue::glue(.names, .formant = ".formant")

  prev_attr <- attributes(.data)$norminfo

  check_grouping(.data, {{ .by }}, call = .call)

  # grouping <- rlang::enquo(.by)
  group_pos <- tidyselect::eval_select(
    enquo(.by),
    data = .data
  )

  target_pos <- try_fetch(
    tidyselect::eval_select(targets, data = .data),
    error = \(cnd) selection_errors(cnd, arg = "...", call = .call)
  )

  # adding an id col for safety in
  # pivoting
  .data <- dplyr::mutate(
    .data,
    .id = dplyr::row_number()
  ) |>
    dplyr::relocate(
      !!sym(".id"),
      .before = 1
    )

  # all normalization happens
  # longwise
  .data <- tidyr::pivot_longer(
    .data,
    any_of(names(target_pos)),
    names_to = ".formant_name",
    values_to = ".formant_orig"
  ) |>
    dplyr::mutate(
      .formant_num = name_to_formant_num(!!sym(".formant_name")),
      .formant = .pre_trans(!!sym(".formant_orig"))
    ) |>
    arrange(
      !!sym(".formant_num")
    )

  # see if the data is grouped
  grouped_by <- dplyr::group_vars(
    .data
  )

  # augment grouping as necessary to
  # match .by_formant
  norm_grouping <- expr({{ .by }})
  if (.by_formant & length(grouped_by > 0)) {
    .data <- dplyr::group_by(
      .data,
      !!sym(".formant_name"),
      .add = TRUE
    )
  } else if (.by_formant) {
    norm_grouping <- expr(c(!!norm_grouping, !!sym(".formant_name")))
  }

  if (.by_token & length(grouped_by > 0)) {
    .data <- dplyr::group_by(
      .data,
      !!sym(".id"),
      .add = TRUE
    )
  } else if (.by_token) {
    norm_grouping <- expr(c(!!norm_grouping, !!sym(".id")))
  }

  .data <- dplyr::mutate(
    .data,
    .by = !!norm_grouping,
    .L = {{ .L }},
    .S = {{ .S }},
    "{.names2}" := .post_trans((!!sym(".formant") - .L) / .S)
  ) |>
    mutate(
      .formant = !!sym(".formant_orig")
    )

  # set up value columns for pivoting
  # back wide
  if (!.keep_params) {
    .data <- dplyr::select(
      .data,
      -c(!!sym(".L"), !!sym(".S"))
    )
  }
  .data <- dplyr::select(
    .data,
    -!!sym(".formant_num"),
    -!!sym(".formant_orig")
  )

  # pivot_back wide
  .data <- tidyr::pivot_wider(
    .data,
    names_from = !!sym(".formant_name"),
    values_from = c(
      tidyselect::starts_with(".formant") & !tidyselect::all_of(".formant_name"),
      tidyselect::any_of(c(".L", ".S"))
    ),
    names_glue = "{.formant_name}_{.value}"
  )

  # move normalized columns adjacent to
  # original
  .data <- .data |>
    dplyr::select(-any_of(".id")) |>
    dplyr::relocate(
      c(
        tidyselect::matches("_.formant"),
        tidyselect::ends_with("_.L"),
        tidyselect::ends_with("_.S")
      ),
      .before = min(target_pos)
    )

  # remove _.col from names
  .data <- dplyr::rename_with(
    .data,
    .cols = tidyselect::matches("_.formant"),
    .fn = \(x) stringr::str_remove(x, "_.formant")
  )

  # if .drop_orig
  if (.drop_orig) {
    .data <- dplyr::select(
      .data,
      -(!!targets)
    )
  }

  norm_info <- list(
    .targets = names(target_pos),
    .norm_cols = glue::glue(.names, .formant = names(target_pos)),
    .by = names(group_pos),
    .by_formant = .by_formant,
    .norm_procedure = "tidynorm::norm_generic",
    .pre_trans = as.character(substitute(.pre_trans)),
    .post_trans = as.character(substitute(.post_trans)),
    .norm = glue::glue("(.formant - {quo_name(enquo(.L))})/({quo_name(enquo(.S))})")
  )

  if (.by_token) {
    norm_info <- c(
      norm_info,
      list(.by_token = .by_token)
    )
  }

  attr(.data, "norminfo") <- prev_attr
  attr(.data, "normalized") <- TRUE
  .data <- append_norm_info(
    .data,
    norm_info
  )

  wrap_up(.data, .silent)

  return(.data)
}

#' Lobanov Normalize
#' @inheritParams norm_generic
#'
#' @param .by_formant Ignored by this procedure
#' @eval options::as_params(".silent" = "tidynorm.silent")
#'
#' @details
#'
#' \deqn{
#'   \hat{F}_{ij} = \frac{F_{ij} - L_i}{S_i}
#' }
#'
#' \deqn{
#'   L_i = \frac{1}{N}\sum_{j=1}^{N}F_{ij}
#' }
#'
#' \deqn{
#'   S_i = \sqrt{\frac{\sum(F_{ij}-L_i)^2}{N-1}}
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data frame of Lobanov normalized formant values.
#'
#' @references
#' Lobanov, B. (1971). Classification of Russian vowels spoken by different listeners.
#' Journal of the Acoustical Society of America, 49, 606–608.
#'
#' @example inst/examples/ex-norm_lobanov.R
#' @importFrom options opt
#' @export
norm_lobanov <- function(
    .data,
    ...,
    .by = NULL,
    .by_formant = TRUE,
    .drop_orig = FALSE,
    .keep_params = FALSE,
    .names = "{.formant}_z",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- rlang::expr(c(...))

  .data <- norm_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .pre_trans = identity,
    .post_trans = identity,
    .L = mean(!!sym(".formant"), na.rm = T),
    .S = sd(!!sym(".formant"), na.rm = T),
    .by_formant = TRUE,
    .drop_orig = .drop_orig,
    .keep_params = .keep_params,
    .names = .names,
    .silent = TRUE
  )

  .data <- update_norm_info(
    .data,
    list(.norm_procedure = "tidynorm::norm_lobanov")
  )

  wrap_up(.data, .silent)

  return(.data)
}

#' Nearey Normalize
#' @inheritParams norm_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#' @importFrom rlang `!!`
#'
#' @details
#' When formant extrinsic:
#' \deqn{
#'  \hat{F}_{ij} = \log(F_{ij}) - L
#' }
#' \deqn{
#'  L = \frac{1}{MN}\sum_{i=1}^M\sum_{j=1}^N \log(F_{ij})
#' }
#'
#' When formant intrinsic:
#' \deqn{
#'  \hat{F}_{ij} = \log(F_{ij}) - L_{i}
#' }
#'
#' \deqn{
#'   L_i = \frac{1}{N}\sum_{j=1}^{N}\log(F_{ij})
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data frame of Nearey normalized formant values.
#'
#' @example inst/examples/ex-norm_nearey.R
#'
#' @references
#' Nearey, T. M. (1978). Phonetic Feature Systems for Vowels \[Ph.D.\].
#' University of Alberta.
#' @export
norm_nearey <- function(
    .data,
    ...,
    .by = NULL,
    .by_formant = FALSE,
    .drop_orig = FALSE,
    .keep_params = FALSE,
    .names = "{.formant}_lm",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- rlang::expr(c(...))

  .data <- norm_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .pre_trans = log,
    .post_trans = identity,
    .L = mean(!!sym(".formant"), na.rm = T),
    .S = 1,
    .by_formant = .by_formant,
    .drop_orig = .drop_orig,
    .keep_params = .keep_params,
    .names = .names,
    .silent = TRUE
  )

  .data <- update_norm_info(
    .data,
    list(.norm_procedure = "tidynorm::norm_nearey")
  )

  wrap_up(.data, .silent)

  return(.data)
}

#' Delta F Normalize
#' @inheritParams norm_generic
#'
#' @param .by_formant Ignored by this procedure
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#' \deqn{
#'  \hat{F}_{ij} = \frac{F_{ij}}{S}
#' }
#' \deqn{
#'  S = \frac{1}{MN}\sum_{i=1}^M\sum_{j=1}^N \frac{F_{ij}}{i-0.5}
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @references
#' Johnson, K. (2020). The \eqn{\Delta}F method of vocal tract length normalization for vowels.
#' Laboratory Phonology: Journal of the Association for Laboratory Phonology, 11(1),
#' Article 1. \doi{10.5334/labphon.196}
#'
#' @returns
#' A data frame of Delta F normalized formant values.
#'
#' @example inst/examples/ex-norm_deltaF.R
#'
#' @export
norm_deltaF <- function(
    .data,
    ...,
    .by = NULL,
    .by_formant = FALSE,
    .drop_orig = FALSE,
    .keep_params = FALSE,
    .names = "{.formant}_df",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- rlang::expr(c(...))

  .data <- norm_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .pre_trans = identity,
    .post_trans = identity,
    .L = 0,
    .S = mean(!!sym(".formant") / (!!sym(".formant_num") - 0.5), na.rm = T),
    .by_formant = FALSE,
    .drop_orig = .drop_orig,
    .keep_params = .keep_params,
    .names = .names,
    .silent = TRUE
  )

  .data <- update_norm_info(
    .data,
    list(.norm_procedure = "tidynorm::norm_deltaF")
  )

  wrap_up(.data, .silent)

  return(.data)
}

#' Watt & Fabricius Normalize
#' @inheritParams norm_generic
#' @param .by_formant Ignored by this procedure
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#' This is a modified version of the Watt & Fabricius Method. The original
#' method identified point vowels over which F1 and F2 centroids were calculated.
#' The procedure here just identifies centroids by taking the mean of
#' all formant values.
#'
#' \deqn{
#' \hat{F}_{ij} = \frac{F_{ij}}{S_i}
#' }
#'
#' \deqn{
#'  S_i = \frac{1}{N}\sum_{j=1}^N F_{ij}
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data fame of Watt & Fabricius normalized formant values.
#'
#' @references
#' Watt, D., & Fabricius, A. (2002). Evaluation of a technique for improving the
#' mapping of multiple speakers’ vowel spaces in the F1 ~ F2 plane.
#' Leeds Working Papers in Linguistics and Phonetics, 9, 159–173.
#'
#' @example inst/examples/ex-norm_wattfab.R
#'
#' @export
norm_wattfab <- function(
    .data,
    ...,
    .by = NULL,
    .by_formant = TRUE,
    .drop_orig = FALSE,
    .keep_params = FALSE,
    .names = "{.formant}_wf",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- rlang::expr(c(...))

  .data <- norm_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .pre_trans = identity,
    .post_trans = identity,
    .L = 0,
    .S = mean(!!sym(".formant"), na.rm = T),
    .by_formant = TRUE,
    .drop_orig = .drop_orig,
    .keep_params = .keep_params,
    .names = .names,
    .silent = TRUE
  )

  .data <- update_norm_info(
    .data,
    list(.norm_procedure = "tidynorm::norm_wattfab")
  )

  wrap_up(.data, .silent)

  return(.data)
}

#' Bark Difference Normalize
#'
#' @inheritParams norm_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#' This is a within-token normalization technique. First all formants are
#' converted to Bark (see [hz_to_bark]), then, within each token, F3 is
#' subtracted from F1 and F2.
#'
#' \deqn{
#' \hat{F}_{ij} = F_{ij} - L_j
#' }
#'
#' \deqn{
#' L_j = F_{3j}
#' }
#'
#' @returns
#' A data frame of Bark Difference normalized formant values
#'
#' @references
#' Syrdal, A. K., & Gopal, H. S. (1986). A perceptual model of vowel
#' recognition based on the auditory representation of American English vowels.
#' The Journal of the Acoustical Society of America, 79(4), 1086–1100.
#' \doi{10.1121/1.393381}
#' @example inst/examples/ex-norm_barkz.R
#' @export
norm_barkz <- function(
    .data,
    ...,
    .by = NULL,
    .drop_orig = FALSE,
    .keep_params = FALSE,
    .names = "{.formant}_bz",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- rlang::expr(c(...))
  target_pos <- tidyselect::eval_select(targets, .data)
  formant_nums <- name_to_formant_num(names(target_pos))

  if (length(target_pos) < 3 ) {
    cli_abort(
      message = c(
        "{.fn tidynorm::norm_barkz} requires F3."
      )
    )
  }

  f3 <- names(target_pos)[formant_nums == 3]

  .data <- norm_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .pre_trans = hz_to_bark,
    .post_trans = identity,
    .L = (!!sym(".formant"))[3],
    .S = 1,
    .by_formant = FALSE,
    .by_token = TRUE,
    .drop_orig = .drop_orig,
    .keep_params = .keep_params,
    .names = .names,
    .silent = TRUE
  )

  .data <- update_norm_info(
    .data,
    list(
      .norm_procedure = "tidynorm::norm_barkz",
      .f3 = f3
    )
  )

  wrap_up(.data, .silent)

  return(.data)
}

test_that(
  "option setting works",
  {
    tidynorm_options(.silent = TRUE, .warnings = FALSE)
    expect_no_message({
      speaker_data |> norm_generic(F1:F3)
    })

    expect_no_warning({
      speaker_data |> norm_generic(F1:F3)
    })

    tidynorm_options(.silent = FALSE, .warnings = TRUE)
    expect_message({
      speaker_data |> norm_generic(F1:F3, .by = speaker)
    })
    expect_warning({
      speaker_data |> norm_generic(F1:F3)
    })
  }
)

test_that(
  "formants are correctly ordered in normalization",
  {
    # Subtract F3 from all formants
    f3_minus <- speaker_data |>
      norm_generic(
        # pass targets in reverse order
        F3:F1,
        .by = speaker,
        .by_token = TRUE,
        .by_formant = FALSE,
        .L = .formant[3],
        .names = "{.formant}_n",
        .silent = T
      )

    # F3 should be zero
    expect_true(
      all(f3_minus$F3_n == 0, na.rm = T)
    )

  }
)


test_that(
  "idiosyncratic formant column names are ok",
  {
    # create some idiosyncratic names
    speaker_data |>
      mutate(
        across(
          F1:F3,
          \(x)x,
          .names = "{.col}_1"
        ),
        across(
          F1:F3,
          \(x)x,
          .names = "{tolower(.col)}"
        ),
        F_one = F1,
        F_two = F2,
        F_three = F3,
        first_formant = F1,
        second_formant = F2,
        third_formant = F3
        ) ->
      weird_names

    # test 1
    norm1 <- weird_names |>
      norm_generic(
        F1_1:F3_1,
        .by = speaker,
        .by_token = TRUE,
        .by_formant = FALSE,
        .L = .formant[3],
        .names = "{.formant}_n",
        .silent = T
      )

    expect_true(
      all(norm1$F3_1_n == 0, na.rm = T)
    )

    # test 2
    norm2 <- weird_names |>
      norm_generic(
        F_one:F_three,
        .by = speaker,
        .by_token = TRUE,
        .by_formant = FALSE,
        .L = .formant[3],
        .names = "{.formant}_n",
        .silent = T
      )

    expect_true(
      all(norm2$F_three_n == 0, na.rm = T)
    )


    # test 3
    norm3 <- weird_names |>
      norm_generic(
        first_formant:third_formant,
        .by = speaker,
        .by_token = TRUE,
        .by_formant = FALSE,
        .L = .formant[3],
        .names = "{.formant}_n",
        .silent = T
      )

    expect_true(
      all(norm3$third_formant_n == 0, na.rm = T)
    )

  }
)

test_that("grouping managed", {
  # group_by() |> norm_ should error
  expect_error({
    speaker_data |>
      dplyr::group_by(speaker) |>
      norm_generic(
        F1:F3,
        .by = speaker
      )
  })

  expect_no_error({
    speaker_data |>
      dplyr::group_by(speaker) |>
      norm_nearey(
        F1:F3,
        .by_formant = T,
        .silent = T
      )
  })
})

test_that("original formants aren't transformed",{
  orig_f1 <- speaker_data$F1
  log_f1 <- log(speaker_data$F1)

  lnorm <- speaker_data |>
    norm_generic(
      F1:F3,
      .by = speaker,
      .pre_trans = log,
      .silent = T,
      .names = "{.formant}_n"
    )

  expect_equal(orig_f1, lnorm$F1, ignore_attr = TRUE)
  expect_equal(log_f1, lnorm$F1_n, ignore_attr = TRUE)
}
)

test_that(
  "point barkz errors on no f3", {
    expect_error(
      speaker_data |> norm_barkz(F1:F2, .by = speaker)
    )

    expect_no_error(
      speaker_data |> norm_barkz(F1:F3, .by = speaker, .silent = T)
    )

    expect_error(
      speaker_data |> norm_barkz(c(F1, F3), .by = speaker, .silent = T)
    )
  }
)


test_that(
  "track barkz errors on no f3", {
    expect_error(
      speaker_tracks |> norm_track_barkz(
        F1:F2,
        .by = speaker,
        .token_id_col = id,
        .time_col = t
      )
    )

    expect_no_error(
      speaker_tracks |> norm_track_barkz(
        F1:F3,
        .by = speaker,
        .token_id_col = id,
        .silent = TRUE,
        .time_col = t
      )
    )

    expect_error(
      speaker_tracks |> norm_track_barkz(
        c(F1, F3),
        .by = speaker,
        .token_id_col = id,
        .silent = TRUE,
        .time_col = t
      )
    )
  }
)

valid_infos <- c(
  ".step", ".norm_procedure", ".targets", ".f3", ".norm_cols",
  ".token_id_col", ".time_col", ".param_col", ".by", ".by_formant",
  ".by_token", ".pre_trans", ".norm", ".post_trans"
)

test_that(
  "Test that norminfo is stored correctly",
  {
    speaker_data |>
      norm_generic(
        F1:F3,
        .by = speaker,
        .L = mean(.formant, na.rm = T),
        .pre_trans = log,
        .post_trans = \(x) exp(x + 6),
        .silent = T
      ) ->
    data

    expected_infos <- c(
      ".by", ".norm", ".pre_trans", ".post_trans",
      ".targets", ".norm_cols", ".norm_procedure",
      ".by_formant"
    )

    last_norm <- attr(data, "norminfo") |>
      dplyr::last()

    expect_in(
      names(last_norm),
      valid_infos
    )

    expect_in(
      expected_infos,
      names(last_norm)
    )

  }
)

#' @eval options::as_params()
#' @name options_params
#'
NULL

#' @eval options::as_roxygen_docs(
#'   desc = stringr::str_c(
#'     sep = " ",
#'     "Options to control the verbosity of tidynorm functions.",
#'     "The convenience function [tidynorm_options()] will set these",
#'     "options within your current session. For a common behavior across",
#'     "R sessions, set the environment variables as described below.",
#'     "If you've silenced informational messages and want to double check",
#'     "what normalization steps have been taken, use [check_norm()]."
#'   )
#' )
NULL

options::define_option(
  option = "tidynorm.silent",
  default = FALSE,
  desc = "Suppress normalization information messages when running a `norm_*()` function.",
  option_name = "tidynorm.silent"
)

options::define_option(
  option = "tidynorm.warnings",
  default = TRUE,
  desc = "Print warnings from tidynorm functions.",
  option_name = "tidynorm.warnings"
)


#' Set tidynorm options
#'
#' Set tidynorm verbosity
#'
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @eval options::as_params(".warnings" = "tidynorm.warnings")
#' @importFrom options opt
#'
#' @seealso [tidynorm::options]
#'
#' @export
#'
#' @examples
#' tidynorm_options(.silent = TRUE, .warnings = FALSE)
#'
#' speaker_data |>
#'   norm_generic(F1:F3) ->
#'   norm1
#'
#' tidynorm_options(.silent = FALSE, .warnings = TRUE)
#'
#' speaker_data |>
#'   norm_generic(F1:F3) ->
#'   norm2
#'
tidynorm_options <- function(
    .silent = opt("tidynorm.silent"),
    .warnings = opt("tidynorm.warnings")
){
  options::opt_set("tidynorm.silent", .silent)
  options::opt_set("tidynorm.warnings", .warnings)
}

library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

speaker_dct <- speaker_tracks |>
  mutate(
    across(
      F1:F3,
      log
    )
  ) |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )

# Normalize DCT coefficients
speaker_dct_norm <- speaker_dct |>
  norm_dct_nearey(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param
  )

# Apply average and apply inverse dct
# to plot tracks
track_norm_means <- speaker_dct_norm |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      ends_with("_lm"),
      mean
    )
  ) |>
  reframe_with_idct(
    ends_with("_lm"),
    .by = speaker,
    .token_id_col = vowel,
    .param_col = .param
  )


if (ggplot2_inst) {
  track_norm_means |>
    ggplot(
      aes(F2_lm, F1_lm, color = speaker)
    ) +
    geom_path(
      aes(
        group = interaction(speaker, vowel)
      )
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}

#' Discrete Cosine Transform
#' @param x A vector or matrix to which the discrete cosine transform is applied
#'
#' @details
#' The DCT definitions here are based on the python `scipy.fft.dct` definitions.
#' Specifically this use:
#'
#' ```python
#' # python code
#' scipy.fft.dct(x, norm = "forward", orthogonalize = True)
#' ```
#'
#' \deqn{
#' y_k = \frac{1}{zN} \sum_{j=0}^{N-1}x_j\cos\left(\frac{\pi k(2j+1)}{2N}\right)
#' }
#'
#' \deqn{
#' z = \begin{cases}
#'    \sqrt{2}& \text{for }k=0\\
#'    1 & \text{for }k>0
#'  \end{cases}
#' }
#'
#'
#' For the Inverse Discrete Cosine Transform, see [idct].
#'
#' @returns
#' Returned value depends on `x`.
#' @examples
#' x <- seq(0, 1, length = 10)
#' y <- 5 + x + (2 * (x^2)) + (-2 * (x^4))
#'
#' dct_coefs <- dct(y)
#'
#' @export
dct <- function(x) {
  UseMethod("dct")
}

#' @rdname dct
#' @returns
#' When passed a numeric vector, returns a numeric
#' vector the same size as `x` with the DCT Coefficients.
#'
#' @method dct numeric
#' @export
#' @keywords internal
dct.numeric <- function(x) {
  coefs <- dct_fun(x, kk = length(x))[, 1]
  return(coefs)
}

#' @rdname dct
#' @export
#' @returns
#' When passed a matrix, returns a matrix
#' the same size as `x` with the DCT Coefficients.
#' @keywords internal
dct.matrix <- function(x) {
  out <- dct_mat(x, kk = nrow(x))
  colnames(out) <- colnames(x)
  out
}

registerS3method("dct", "numeric", method = dct.numeric)
registerS3method("dct", "matrix", method = dct.matrix)

#' DCT Basis
#'
#' The Discrete Cosine Transform basis functions
#'
#' @param n The length of the basis.
#' @param k The number of basis functions.
#'
#' @details
#' This function will generate the DCT basis functions.
#'
#' @returns
#' A \eqn{n\times k} matrix
#'
#' @examples
#' basis <- dct_basis(100, 5)
#' matplot(basis, type = "l", lty = 1)
#'
#' @export
dct_basis <- function(n, k) {
  cos_bank(n, k)
}

#' Inverse Discrete Cosine Transform
#'
#' The Inverse DCT
#'
#' @param y A vector or matrix of DCT coefficients
#' @param n The desired length of the idct
#'
#' @details
#' Applies the Inverse DCT (see [dct] for more details).
#'
#' \deqn{
#' x_j = \sqrt{2}y_0 + 2\sum_{k=1}^{N-1} y_k \cos\left(\frac{\pi k(2j+1)}{2J}\right)
#' }
#'
#' @returns
#' The returned value depends on the values in `y`.
#'
#'
#' @examples
#' x <- seq(0, 1, length = 10)
#' y <- 5 + x + (2 * (x^2)) + (-2 * (x^4))
#'
#' dct_coefs <- dct(y)
#' recovered_y <- idct(dct_coefs)
#'
#' plot(y, recovered_y)
#'
#' @export
idct <- function(y, n) {
  UseMethod("idct")
}

#' @rdname idct
#' @returns When passed a numeric vector, returns numeric vector of length `n`.
#' @export
#' @method idct numeric
#' @keywords internal
idct.numeric <- function(y, n = length(y)) {
  x <- idct_fun(y, n = n)[, 1]
  return(x)
}

#' @rdname idct
#' @returns
#' When passed a matrix, returns a matrix with `n` rows and the same number of columns as `y`.
#' @export
#' @method idct matrix
#' @keywords internal
idct.matrix <- function(y, n = nrow(y)) {
  x <- idct_mat(y, n = n)
  colnames(x) <- colnames(y)
  return(x)
}

registerS3method("idct", "numeric", method = idct.numeric)
registerS3method("idct", "matrix", method = idct.matrix)

#' Inverse Discrete Cosine Transform Rate
#'
#' The first derivative of the Inverse DCT
#'
#' @param y DCT coefficients
#' @param n The desired length of the idct
#'
#' @details
#' Returns the first derivative (rate of change) of
#' the Inverse DCT (see [dct] for more details).
#'
#' \deqn{
#' \frac{\delta x_j}{\delta j} = -2\frac{\pi k}{J}\sum_{k=1}^{N-1} y_k \sin\left(\frac{\pi k(2j+1)}{2J}\right)
#' }
#'
#' @returns
#' A vector with the first derivative
#' of the inverse DCT
#'
#'
#' @examples
#' x <- seq(0, 1, length = 10)
#' y <- 5 + x + (2 * (x^2)) + (-2 * (x^4))
#'
#' dct_coefs <- dct(y)
#' y_rate <- idct_rate(dct_coefs)
#'
#' plot(y)
#' plot(y_rate)
#'
#' @export
idct_rate <- function(y, n = length(y)) {
  x <- idct_prime(y, n = n)
  return(x)
}

#' Inverse Discrete Cosine Transform Acceleration
#'
#' The second derivative of the Inverse DCT
#'
#' @param y DCT coefficients
#' @param n The desired length of the idct
#'
#' @details
#' Returns the second derivative (acceleration) of
#' the Inverse DCT (see [dct] for more details).
#'
#' \deqn{
#' \frac{\delta^2 x_j}{\delta j^2} = -2\left(\frac{\pi k}{J}\right)^2\sum_{k=1}^{N-1} y_k \cos\left(\frac{\pi k(2j+1)}{2J}\right)
#' }
#'
#' @returns
#' A vector with the second derivative
#' of the inverse DCT
#'
#'
#' @examples
#' x <- seq(0, 1, length = 10)
#' y <- 5 + x + (2 * (x^2)) + (-2 * (x^4))
#'
#' dct_coefs <- dct(y)
#' y_accel <- idct_accel(dct_coefs)
#'
#' plot(y)
#' plot(y_accel)
#'
#' @export
idct_accel <- function(y, n = length(y)) {
  x <- idct_dprime(y, n = n)
  return(x)
}

#' Reframe with DCT
#'
#' Reframe data columns using the Discrete Cosine Transform
#'
#' @param .data A data frame
#' @param ... [`<tidy-select>`][dplyr::dplyr_tidy_select] One or more unquoted
#' expressions separated by commas. These should target the vowel formant.
#' @param .by [`<tidy-select>`][dplyr::dplyr_tidy_select] A grouping column.
#' @param .token_id_col [`<tidy-select>`][dplyr::dplyr_tidy_select] The token ID column.
#' @param .time_col A time column.
#' @param .order The number of DCT parameters to return. If `NA`, all DCT
#' parameters will be returned.
#'
#' @details
#' This function will tidily apply the Discrete Cosine Transform with forward
#' normalization (see [dct] for more info) to the targeted columns.
#'
#' ### Identifying tokens
#' The DCT only works on a by-token basis, so there must be a column that
#' uniquely identifies (or, in combination with a `.by` grouping, uniquely
#' identifies) each individual token. This column should be passed to
#' `.token_id_col`.
#'
#' ### Order
#' The number of DCT coefficients to return is defined by `.order`. The default
#' value is 5. Larger numbers will lead to less smoothing when the Inverse
#' DCT is applied (see [idct]). Smaller numbers will lead to more smoothing.
#'
#' If `NA` is passed to `.order`, all DCT parameters will be returned, which
#' when the Inverse DCT is supplied, will completely reconstruct the original
#' data.
#'
#' ### Sorting by Time
#' An optional `.time_col` can also be defined to ensure that the data is
#' correctly arranged by time.
#'
#' @returns
#' A data frame with with the targeted DCT coefficients, along with two
#' additional columns
#'
#' \describe{
#'  \item{.param}{The nth DCT coefficient number}
#'  \item{.n}{The number of original data values}
#' }
#'
#' @example inst/examples/ex-reframe_with_dct.R
#' @export
reframe_with_dct <- function(
    .data,
    ...,
    .token_id_col = NULL,
    .by = NULL,
    .time_col = NULL,
    .order = 5) {
  targets <- expr(...)
  target_pos <- tidyselect::eval_select(targets, .data)

  cols <- enquos(
    .token_id_col = .token_id_col,
    .time = .time_col,
    .by = .by
  )

  for (col in cols) {
    try_fetch(
      tidyselect::eval_select(col, data = .data),
      error = \(cnd) selection_errors(cnd)
    )
  }

  order <- if (!is.finite(.order)) {
    expr(dplyr::n())
  } else {
    expr(.order)
  }

  # make sure groupings are ok
  check_grouping(.data, {{ .by }})


  if (quo_is_null(enquo(.time_col)) & options::opt("tidynorm.warnings")) {
    cli_par()
    cli_inform(
      c(
        "i" = "No {.arg .time_col} provided.",
        "i" = "Assuming {.arg .data} is arranged by time."
      )
    )
    cli_end()
  } else {
    .data <- dplyr::arrange(.data, {{ .time_col }}) |>
      dplyr::select(-{{ .time_col }})
  }

  grouping_list <- make_dct_grouping(
    .data,
    {{ .by }},
    {{ .token_id_col }}
  )
  .data <- grouping_list$.data
  by_grouping <- grouping_list$by_grouping
  joining <- grouping_list$joining

  orig <- dplyr::select(
    .data,
    -!!targets
  ) |>
    dplyr::slice(
      .by = !!by_grouping,
      1
    )

  dct_df <- .data |>
    dplyr::mutate(
      .by = !!by_grouping,
      dplyr::across(
        !!targets,
        \(x){
          if (mean(is.finite(x)) < 0.9) {
            x <- NA
          }
          return(x)
        }
      )
    ) |>
    dplyr::reframe(
      .by = !!by_grouping,
      .param = (1:!!order) - 1,
      dplyr::across(
        !!targets,
        \(x) dct(x)[1:!!order]
      ),
      .n = dplyr::n()
    )

  out_df <- dplyr::left_join(
    orig,
    dct_df,
    by = unique(joining)
  ) |>
    dplyr::relocate(
      !!targets,
      .before = min(target_pos)
    )

  return(out_df)
}

#' Reframe with IDCT
#'
#' Reframe data columns using the Inverse Discrete Cosine Transform
#'
#' @inheritParams reframe_with_dct
#' @param .param_col A column identifying the DCT parameter number
#' @param .n The size of the outcome of the IDCT
#' @param .rate Whether or not to include the rate of change of signal.
#' @param .accel Whether or not to include acceleration of signal.
#'
#' @details
#' This will apply the Inverse Discrete Cosine Transform to the targeted
#' columns. See [idct].
#'
#' ### Identifying tokens
#' The IDCT only works on a by-token basis, so there must be a column that
#' uniquely identifies (or, in combination with a `.by` grouping, uniquely
#' identifies) each individual token. This column should be passed to
#' `.token_id_col`.
#'
#' ### Size of the output
#' The output of the IDCT can be arbitrarily long as defined by the `.n`
#' argument. `.n` can either be an integer, or an unqoted data column.
#'
#' ### The Parameter Column
#' The order of the DCT parameters is crucially important. The optional
#' `.param_col` will ensure the data is properly arranged.
#'
#' @returns
#' A data frame with the IDCT of the targeted columns along with an
#' additional `.time` column.
#'
#' \describe{
#'  \item{.time}{A column from 1 to `.n` by token}
#' }
#'
#' @example inst/examples/ex-reframe_with_idct.R
#'
#' @export
reframe_with_idct <- function(
    .data,
    ...,
    .token_id_col = NULL,
    .by = NULL,
    .param_col = NULL,
    .n = 20,
    .rate = FALSE,
    .accel = FALSE) {
  targets <- expr(c(...))
  cols <- enquos(
    .token_id_col = .token_id_col,
    .param_col = .param_col,
    .by = .by
  )

  target_pos <- try_fetch(
    tidyselect::eval_select(targets, data = .data),
    error = \(cnd) selection_errors(cnd)
  )

  for (col in cols) {
    try_fetch(
      tidyselect::eval_select(col, data = .data),
      error = \(cnd) selection_errors(cnd)
    )
  }

  # make sure groupings are ok
  check_grouping(.data, {{ .by }})

  if (quo_is_null(cols$.param_col) & options::opt("tidynorm.warnings")) {
    cli_par()
    cli_inform(
      c(
        "i" = "No {.arg .param_col} provided.",
        "i" = "Assuming {.arg .data} is arranged by parameter."
      )
    )
    cli_end()
  } else {
    .data <- dplyr::arrange(.data, {{ .param_col }}) |>
      dplyr::select(-{{ .param_col }})
  }

  grouping_list <- make_dct_grouping(
    .data,
    {{ .by }},
    {{ .token_id_col }}
  )
  .data <- grouping_list$.data
  by_grouping <- grouping_list$by_grouping
  joining <- grouping_list$joining

  orig <- dplyr::select(
    .data,
    -!!targets
  ) |>
    dplyr::slice(
      .by = !!by_grouping,
      1
    )

  if (!.rate & !.accel) {
    idct_df <- .data |>
      dplyr::reframe(
        .by = !!by_grouping,
        .time = 1:dplyr::first({{ .n }}),
        dplyr::across(
          !!targets,
          \(x) idct(x, n = {{ .n }}[1])
        )
      )
  } else {
    idct_operation <- list(
      s = \(x) idct(x, n = {{ .n }}[1])
    )
    if (.rate) {
      idct_operation$rate <- \(x) idct_prime(x, {{ .n }}[1])
    }
    if (.accel) {
      idct_operation$accel <- \(x) idct_dprime(x, n = {{ .n }}[1])
    }

    idct_df <- .data |>
      dplyr::reframe(
        .by = !!by_grouping,
        .time = 1:({{ .n }}[1]),
        dplyr::across(
          !!targets,
          idct_operation
        )
      )
  }

  out_df <- dplyr::left_join(
    orig,
    idct_df,
    by = unique(joining)
  ) |>
    relocate(
      tidyselect::starts_with(names(target_pos)),
      .before = min(target_pos)
    )
  return(out_df)
}

#' Reframe with DCT Smooth
#'
#' Apply a DCT Smooth to the targeted data
#'
#' @inheritParams reframe_with_dct
#' @inheritParams reframe_with_idct
#'
#' @details
#' This is roughly equivalent to applying [reframe_with_dct] followed by
#' [reframe_with_idct]. As long as the value passed to `.order` is less than
#' the length of the each token's data, this will result in a smoothed version
#' of the data.
#'
#' ### Identifying tokens
#' The DCT only works on a by-token basis, so there must be a column that
#' uniquely identifies (or, in combination with a `.by` grouping, uniquely
#' identifies) each individual token. This column should be passed to
#' `.token_id_col`.
#'
#' ### Order
#' The number of DCT coefficients to return is defined by `.order`. The default
#' value is 5. Larger numbers will lead to less smoothing when the Inverse
#' DCT is applied (see [idct]). Smaller numbers will lead to more smoothing.
#'
#' If `NA` is passed to `.order`, all DCT parameters will be returned, which
#' when the Inverse DCT is supplied, will completely reconstruct the original
#' data.
#'
#' ### Sorting by Time
#' An optional `.time_col` can also be defined to ensure that the data is
#' correctly arranged by time.
#'
#' Additionally, if `.time_col` is provided, the original time column will
#' be included in the output
#'
#' @returns
#' A data frame where the target columns have been smoothed using the
#' DCT, as well as the signal rate of change and acceleration,
#' if requested.
#'
#' @example inst/examples/ex-reframe_with_dct_smooth.R
#'
#'
#' @export
reframe_with_dct_smooth <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .time_col = NULL,
    .order = 5,
    .rate = FALSE,
    .accel = FALSE) {
  targets <- expr(...)
  cols <- enquos(
    .token_id_col = .token_id_col,
    .time_col = .time_col,
    .by = .by
  )

  target_pos <- try_fetch(
    tidyselect::eval_select(targets, data = .data),
    error = \(cnd) selection_errors(cnd)
  )

  for (col in cols) {
    try_fetch(
      tidyselect::eval_select(col, data = .data),
      error = \(cnd) selection_errors(cnd)
    )
  }

  # make sure groupings are ok
  check_grouping(.data, {{ .by }})

  grouping_list <- make_dct_grouping(
    .data,
    {{ .by }},
    {{ .token_id_col }}
  )

  .data <- grouping_list$.data
  by_grouping <- grouping_list$by_grouping
  joining <- grouping_list$joining

  if (!quo_is_null(cols$.time_col)) {
    .time_data <- dplyr::arrange(.data, {{ .time_col }}) |>
      dplyr::select(
        {{ .by }},
        {{ .token_id_col }},
        {{ .time_col }},
        dplyr::group_cols()
      ) |>
      dplyr::mutate(
        .by = !!by_grouping,
        .row = dplyr::row_number()
      )
  }

  .dct_data <- reframe_with_dct(
    .data,
    !!targets,
    .token_id_col = {{ .token_id_col }},
    .by = !!by_grouping,
    .time_col = {{ .time_col }},
    .order = .order
  )

  .dct_smooth <- reframe_with_idct(
    .dct_data,
    !!targets,
    .token_id_col = {{ .token_id_col }},
    .by = !!by_grouping,
    .n = !!sym(".n"),
    .param_col = !!sym(".param"),
    .rate = .rate,
    .accel = .accel
  )

  if (!quo_is_null(cols$.time_col)) {
    .dct_smooth <- dplyr::select(
      .dct_smooth,
      -!!sym(".time")
    ) |>
      dplyr::mutate(
        .by = !!by_grouping,
        .row = dplyr::row_number()
      ) |>
      dplyr::left_join(
        .time_data,
        by = unique(c(joining, ".row"))
      )
  }

  return(.dct_smooth)
}

# generate a curve
x <- seq(0, 1, length = 10)
y <- 5 + x + (2 * (x^2)) + (-2 * (x^4))

test_that("dct coefs from package are equal to established library", {
  # Only run test if scipy is available
  skip_on_cran()
  skip_if(!rlang::is_installed("reticulate"))
  reticulate::py_config()
  skip_if(!reticulate::py_available())
  skip_if(!"scipy" %in% reticulate::py_list_packages()$package)

  # import scipy dct implementation
  scipy <- reticulate::import("scipy")
  py_dct <- scipy$fft$dct

  # get scipy coefficients & strip dim() to match R vector
  py_coefs <- py_dct(y, norm = "forward", orthogonalize = TRUE)
  dim(py_coefs) <- NULL

  # package dct
  r_coefs <- tidynorm::dct(y)

  expect_equal(length(r_coefs), length(py_coefs))
  expect_equal(r_coefs, py_coefs)
})


test_that("idct recovery", {
  coefs <- dct(y)
  recovered <- idct(coefs)

  expect_equal(recovered, y)
})

test_that("dct matrix method", {
  mat <- cbind(y, y / 3)
  coefs <- dct(mat)

  expect_equal(
    dim(coefs),
    dim(mat)
  )

  expect_equal(
    dimnames(coefs),
    dimnames(mat)
  )
})

test_that("idct matrix method", {
  mat <- cbind(y, y / 3)
  coefs <- dct(mat)
  recovered <- idct(coefs)

  expect_equal(
    recovered,
    mat
  )
})


test_that("dct-basis works", {
  rows <- 100
  cols <- 5
  basis <- dct_basis(rows, cols)

  expect_equal(
    nrow(basis),
    rows
  )

  expect_equal(
    ncol(basis),
    cols
  )

  expect_equal(
    basis[1, 1],
    sqrt(2)
  )
})

test_that("reframe with dct works", {
  test_order <- 5

  n_tokens <- speaker_tracks |>
    dplyr::count(
      speaker, id
    ) |>
    nrow()

  speaker_dct1 <- reframe_with_dct(
    .data = speaker_tracks,
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .order = test_order
  )

  speaker_dct2 <- reframe_with_dct(
    .data = speaker_tracks,
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .order = NA
  )

  expect_equal(
    nrow(speaker_dct1),
    n_tokens * test_order
  )

  expect_equal(
    nrow(speaker_dct2),
    nrow(speaker_tracks)
  )

  expect_contains(
    colnames(speaker_dct1),
    c("F1", "F2", "F3")
  )
})

test_that("reframe with idct works", {
  out_len <- 20

  n_tokens <- speaker_tracks |>
    dplyr::count(
      speaker, id
    ) |>
    nrow()

  speaker_dct <- reframe_with_dct(
    .data = speaker_tracks,
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )

  speaker_idct <- reframe_with_idct(
    .data = speaker_dct,
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param,
    .n = out_len
  )

  expect_equal(
    nrow(speaker_idct),
    n_tokens * out_len
  )

  orig_means <- speaker_tracks |>
    dplyr::summarise(
      .by = c(speaker, id),
      dplyr::across(
        F1:F3,
        mean
      )
    ) |>
    tidyr::pivot_longer(
      F1:F3,
      names_to = "formant",
      values_to = "mean"
    )

  idct_ranges <- speaker_idct |>
    dplyr::summarise(
      .by = c(speaker, id),
      dplyr::across(
        F1:F3,
        .fns = list(min = min, max = max)
      )
    ) |>
    tidyr::pivot_longer(
      F1_min:F3_max,
      names_to = c("formant", "metric"),
      names_sep = "_"
    ) |>
    tidyr::pivot_wider(
      names_from = metric,
      values_from = value
    )

  orig_idct_compare <- orig_means |>
    dplyr::left_join(
      idct_ranges
    )

  expect_true(
    all(
      orig_idct_compare$mean > orig_idct_compare$min,
      na.rm = T
    )
  )

  expect_true(
    all(
      orig_idct_compare$mean < orig_idct_compare$max,
      na.rm = T
    )
  )
})

are_names_aligned <- function(orig_cols, new_cols){

  orig_check <- seq_along(orig_cols) |>
    set_names(orig_cols)
  new_check <- seq_along(new_cols) |>
    set_names(new_cols)

  for (i in seq_along(orig_check)) {
    if (i == 1) {
      expect_equal(
        names(new_check)[i], names(orig_check)[i]
      )
    } else {
      curr_name <- names(orig_check)[i]
      prev_name <- names(orig_check)[i - 1]

      expect_gt(
        new_check[curr_name], new_check[prev_name]
      )
    }
  }
}


test_that(
  "point norm output columns are correctly ordered",
  {
    orig_cols <- colnames(speaker_data)
    norm_data <- speaker_data |> norm_generic(F1:F3, .by = speaker, .silent = T)
    new_cols <- colnames(norm_data)
    are_names_aligned(orig_cols, new_cols)

    formant_start <- speaker_data |>
      dplyr::relocate(
        F1:F3,
        .before = 1
      )

    orig_cols <- colnames(formant_start)
    norm_data <- formant_start |> norm_generic(F1:F3, .by = speaker, .silent = T)
    new_cols <- colnames(norm_data)
    are_names_aligned(orig_cols, new_cols)

  }
)


test_that(
  "track norm output columns are correctly ordered",
  {
    orig_cols <- colnames(speaker_tracks)
    # with time col
    norm_data <- speaker_tracks |> norm_track_generic(
      F1:F2,
      .by = speaker,
      .token_id_col = id,
      .time_col = t,
      .silent = TRUE
      )
    new_cols <- colnames(norm_data)
    are_names_aligned(orig_cols, new_cols)


    # without time col
    norm_data <- speaker_tracks |> norm_track_generic(
      F1:F2,
      .by = speaker,
      .token_id_col = id,
      .silent = TRUE
    )
    new_cols <- colnames(norm_data)
    are_names_aligned(orig_cols, new_cols)

    # formant first with time col
    formant_first <- speaker_tracks |>
      relocate(
        F1:F3, .before = 1
      )

    orig_cols <- colnames(formant_first)
    norm_data <- formant_first |> norm_track_generic(
      F1:F2,
      .by = speaker,
      .token_id_col = id,
      .time_col = t,
      .silent = TRUE
    )
    new_cols <- colnames(norm_data)
    are_names_aligned(orig_cols, new_cols)

    # formant first without time col
    norm_data <- formant_first |> norm_track_generic(
      F1:F2,
      .by = speaker,
      .token_id_col = id,
      .silent = TRUE
    )
    new_cols <- colnames(norm_data)
    are_names_aligned(orig_cols, new_cols)

  }


)

test_that(
  "Relocation with rate & acceleration works",
  {
    new_smooth <- speaker_tracks |>
      dplyr::relocate(
        F1:F3,
        .before = 3
      ) |>
      reframe_with_dct(
        F1:F3,
        .time_col = t,
        .token_id_col = id,
        .by = speaker) |>
      reframe_with_idct(
        F1:F3,
        .param_col = .param,
        .token_id_col = id,
        .by = speaker,
        .rate = T,
        .accel = T
      )

    new_names <- colnames(new_smooth)
    expect_equal(
      which(new_names == "F1_s"),
      3
    )

  }
)

test_that(
  "reframe with idct smooth returns correct lengths",
  {
    speaker_tracks |>
      dplyr::mutate(
        .by = c(speaker, id),
        rowid = dplyr::row_number()
      ) |>
      dplyr::filter(
        (id %% 2 == 0 & rowid %% 2 == 1) |
          id %% 2 == 1
      ) ->
      diff_size

    diff_size |>
      reframe_with_dct_smooth(
        F1:F3,
        .time_col = t,
        .token_id_col = id,
        .by = speaker
      ) ->
      diff_size_smooths

    diff_size |>
      dplyr::count(speaker, id) |>
      dplyr::rename(orig_n = n) ->
      orig_count

    diff_size_smooths |>
      dplyr::count(speaker, id) |>
      dplyr::rename(smooth_n = n) ->
      smooth_count

    orig_count |>
      dplyr::left_join(smooth_count)->
      count_comp

    expect_equal(count_comp$orig_n, count_comp$smooth_n, ignore_attr = TRUE)

  }
)

#' @import rlang
#' @import cli
#' @importFrom stats sd
#' @import dplyr
#' @keywords internal
#' ## usethis namespace: start
#' @importFrom Rcpp sourceCpp
#' @importFrom Rcpp evalCpp
#' @useDynLib tidynorm, .registration = TRUE
## usethis namespace: end
"_PACKAGE"

#' Generic Formant Track Normalization Procedure
#'
#' Normalize formant tracks using Discrete Cosine Transform normalization
#'
#' @param .data A data frame containing vowel formant data
#' @param ... [`<tidy-select>`][dplyr::dplyr_tidy_select] One or more unquoted
#' expressions separated by commas. These should target the vowel formant
#' data columns.
#' @param .by  [`<tidy-select>`][dplyr::dplyr_tidy_select] A selection of
#' columns to group by. Typically a column of speaker IDs.
#' @param .token_id_col [`<data-masking>`][rlang::args_data_masking] A column
#' that identifies token ids.
#' @param .by_formant Whether or not the normalization method is formant
#' intrinsic.
#' @param .by_token Whether or not the normalization method is token
#' intrinsic
#' @param .L An expression defining the location parameter.
#' See Details for more information.
#' @param .S An expression defining the scale parameter.
#' See Details for more information.
#' @param .pre_trans A function to apply to formant values before normalization.
#' @param .post_trans A function to apply to formant values after normalization.
#' @param .time_col [`<data-masking>`][rlang::args_data_masking] A time column.
#' (optional)
#' @param .order The number of DCT parameters to use.
#' @param .names A [glue::glue()] expression for naming the normalized
#' data columns. The `"{.formant}"` portion corresponds to the name of the original
#' formant columns.
#' @param .drop_orig Should the originally targeted columns be dropped.
#' @param .return_dct Whether or not the normalized DCT coefficients themselves
#' should be returned.
#' @param .silent Whether or not the informational message should be printed.
#' @param .call Used for internal purposes.
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#' The following `norm_track_*` procedures were built on top of
#' `norm_track_generic`.
#'
#' - [norm_track_lobanov]
#' - [norm_track_nearey]
#' - [norm_track_deltaF]
#' - [norm_track_wattfab]
#' - [norm_track_barkz]
#'
#' This will normalize vowel formant tracks in the following steps:
#'
#' 1. Any `.pre_trans` transformations will be applied to the formant data.
#' 2. The Discrete Cosine Transform will be applied to the formant data.
#' 3. Location `.L` and Scale `.S` expressions will be used to summarize the zero<sup>th</sup>
#' DCT coefficients.
#' 4. These location and scale will be used to normalize the DCT coefficients.
#' 5. If `.return_dct = TRUE`, these normalized DCT coefficients will be returned.
#' Otherwise, the Inverse Discrete Cosine Transform will be applied to the
#' normalized DCT coefficients.
#' 6. Any `.post_trans` transformations will be applied.
#'
#' ### Location and Scale expressions
#' All normalization procedures built on [norm_track_generic] work by normalizing
#' DCT coefficients directly. If \eqn{F_k} is the k<sup>th</sup> DCT coefficient
#' the normalization procedure is
#'
#' \deqn{
#' \hat{F}_k = \frac{F_k - L'}{\sqrt{2}S}
#' }
#' \deqn{
#' L' = \begin{cases}
#'    L & \text{for }k=0\\
#'    0 & \text{for }k>0
#'  \end{cases}
#' }
#'
#' Rather than requiring users to remember to multiply expressions for \eqn{S}
#' by \eqn{\sqrt{2}}, this is done by [norm_track_generic] itself, to allow greater
#' parallelism with how [norm_generic] works.
#'
#' **Note**: If you want to scale values by a constant in the normalization,
#' you'll need to divide the constant by `sqrt(2)`. Post-normalization scaling
#' (e.g. re-scaling to formant-like values) is probably best handled with a
#' function passed to `.post_trans`.
#'
#' The expressions for calculating \eqn{L} and \eqn{S} can be
#' passed to `.L` and `.S`, respectively. Available values for
#' these expressions are
#'
#' \describe{
#'  \item{`.formant`}{The original formant value}
#'  \item{`.formant_num`}{The number of the formant. (e.g. 1 for F1, 2 for F2 etc)}
#' }
#'
#' Along with any data columns from your original data.
#'
#' ### Identifying tokens
#' Track normalization requires identifying individual tokens, so there must be a column that
#' uniquely identifies (or, in combination with a `.by` grouping, uniquely
#' identifies) each individual token. This column should be passed to
#' `.token_id_col`.
#'
#' ### Order
#' The number of DCT coefficients used is defined by `.order`. The default
#' value is 5. Larger numbers will lead to less smoothing, and smaller numbers
#' will lead to more smoothing.
#'
#' @returns
#' A data frame of normalized formant tracks.
#'
#' @example inst/examples/ex-norm_track_generic.R
#'
#' @export
norm_track_generic <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .by_formant = FALSE,
    .by_token = FALSE,
    .time_col = NULL,
    .L = 0,
    .S = 1 / sqrt(2),
    .pre_trans = identity,
    .post_trans = identity,
    .order = 5,
    .return_dct = FALSE,
    .drop_orig = FALSE,
    .names = "{.formant}_n",
    .silent = opt("tidynorm.silent"),
    .call = caller_env()) {
  if (env_name(.call) == "global") {
    .call2 <- current_env()
  }
  ## Argument checks
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls, .call2)

  # preserve incoming attributes
  prev_attr <- attributes(.data)$norminfo

  # stand-in names
  .names2 <- glue::glue(.names, .formant = ".formant")

  # position info
  targets <- expr(c(...))
  target_pos <- tidyselect::eval_select(targets, .data)
  time_pos <- tidyselect::eval_select(expr({{ .time_col }}), .data)

  cols <- enquos(
    .by = .by,
    .token_id_col = .token_id_col,
    .time_col = .time_col
  )

  for (x in cols) {
    try_fetch(
      tidyselect::eval_select(x, data = .data),
      error = \(cnd) selection_errors(cnd)
    )
  }

  check_grouping(.data, {{ .by }}, .call)

  by_pos <- tidyselect::eval_select(enquo(.by), .data)

  grouping_list <- make_dct_grouping(
    .data,
    {{ .by }},
    {{ .token_id_col }}
  )
  .data <- grouping_list$.data
  by_grouping <- grouping_list$by_grouping
  joining <- grouping_list$joining

  .orig_data <- dplyr::arrange(.data, {{ .time_col }}) |>
    dplyr::select(
      {{ .by }},
      {{ .token_id_col }},
      {{ .time_col }},
      any_of(names(target_pos)),
      dplyr::group_cols()
    ) |>
    dplyr::mutate(
      .by = !!by_grouping,
      .row = dplyr::row_number()
    )

  .data <- dplyr::mutate(
    .data,
    .by = !!by_grouping,
    .row = dplyr::row_number()
  )


  .dct_data <- .data |>
    dplyr::mutate(
      across(
        !!targets,
        .fns = \(x) .pre_trans(x)
      )
    ) |>
    reframe_with_dct(
      !!targets,
      .token_id_col = {{ .token_id_col }},
      .by = !!by_grouping,
      .time_col = {{ .time_col }},
      .order = .order
    )

  .dct_data <- append_norm_info(
    .dct_data,
    list(
      .step = "DCT applied to tracks",
      .token_id_col = quo_name(enquo(.token_id_col)),
      .time_col = quo_name(enquo(.time_col)),
      .by = names(by_pos),
      .token_id_col = quo_name(enquo(.token_id_col))
    )
  )

  normed <- norm_dct_generic(
    .dct_data,
    !!targets,
    .token_id_col = {{ .token_id_col }},
    .by = {{ .by }},
    .by_formant = .by_formant,
    .by_token = .by_token,
    .param_col = !!sym(".param"),
    .L = {{ .L }},
    .S = {{ .S }},
    .names = .names,
    .silent = TRUE,
    .drop_orig = .drop_orig,
    .call = current_env()
  )

  if (.return_dct) {
    normed <- normed |>
      dplyr::rename_with(
        .fn = \(x) stringr::str_remove(x, "_.formant")
      )
    return(normed)
  }

  normed_track <- reframe_with_idct(
    normed,
    matches("_.formant"),
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .param_col = !!sym(".param"),
    .n = !!sym(".n")
  ) |>
    dplyr::mutate(
      dplyr::across(
        matches("_.formant"),
        .fns = \(x) .post_trans(x)
      )
    )

  normed_track <- append_norm_info(
    normed_track,
    list(
      .step = "Inverse DCT applied",
      .param_col = ".param",
      .by = names(by_pos),
      .token_id_col = quo_name(enquo(.token_id_col))
    )
  )

  normed_track <- normed_track |>
    dplyr::relocate(
      dplyr::contains("_.formant"),
      .before = min(c(target_pos, time_pos))
    ) |>
    dplyr::rename_with(
      .fn = \(x) stringr::str_remove(x, "_.formant")
    )


  normed_track <- normed_track |>
    select(-tidyselect::any_of(names(target_pos))) |>
    dplyr::mutate(
      .by = !!by_grouping,
      .row = dplyr::row_number()
    ) |>
    dplyr::left_join(
      .orig_data,
      by = c(joining, ".row")
    ) |>
    dplyr::select(
      -any_of(c(".row", ".n"))
    )

  if (!quo_is_null(cols$.time_col)) {
    normed_track <- normed_track |>
      dplyr::select(
        -sym(".time")
      )
    if (time_pos > max(target_pos)) {
      time_pos <- time_pos + length(target_pos)
    }
  }

  normed_track <- normed_track |>
    dplyr::relocate(
      any_of(names(target_pos)),
      .before = any_of(min(c(target_pos, time_pos)))
    ) |>
    dplyr::relocate(
      {{ .time_col }},
      .before = any_of(time_pos)
    )


  attr(normed_track, "norminfo") <- prev_attr

  norm_info <- list(
    .norm_procedure = "tidynorm::norm_track_generic",
    .targets = names(target_pos),
    .norm_cols = glue::glue(.names, .formant = names(target_pos)),
    .by = names(by_pos),
    .token_id_col = quo_name(enquo(.token_id_col)),
    .by_formant = .by_formant,
    .norm = glue::glue("(.formant - {quo_name(enquo(.L))})/{quo_name(enquo(.S))}"),
    .pre_trans = as.character(substitute(.pre_trans)),
    .post_trans = as.character(substitute(.post_trans))
  )

  if (.by_token) {
    norm_info <- c(
      norm_info,
      list(.by_token = .by_token)
    )
  }

  if (!quo_is_null(cols$.time_col)) {
    norm_info <- c(
      norm_info,
      list(.time_col = quo_name(enquo(.time_col)))
    )
  }

  normed_track <- append_norm_info(
    normed_track,
    norm_info
  )

  wrap_up(normed_track, .silent)

  return(normed_track)
}

#' Lobanov Track Normalization
#'
#' @inheritParams norm_track_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#'
#' \deqn{
#'   \hat{F}_{ij} = \frac{F_{ij} - L_i}{S_i}
#' }
#'
#' \deqn{
#'   L_i = \frac{1}{N}\sum_{j=1}^{N}F_{ij}
#' }
#'
#' \deqn{
#'   S_i = \sqrt{\frac{\sum(F_{ij}-L_i)^2}{N-1}}
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data frame of Lobanov normalized formant tracks.
#'
#' @references
#' Lobanov, B. (1971). Classification of Russian vowels spoken by different listeners.
#' Journal of the Acoustical Society of America, 49, 606–608.
#'
#' @example inst/examples/ex-norm_track_lobanov.R
#' @export
norm_track_lobanov <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .time_col = NULL,
    .order = 5,
    .return_dct = FALSE,
    .drop_orig = FALSE,
    .names = "{.formant}_z",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- expr(...)
  normed <- norm_track_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .by_formant = TRUE,
    .L = mean(!!sym(".formant"), na.rm = T),
    .S = sd(!!sym(".formant"), na.rm = T),
    .pre_trans = identity,
    .post_trans = identity,
    .time_col = {{ .time_col }},
    .order = .order,
    .return_dct = .return_dct,
    .drop_orig = .drop_orig,
    .names = .names,
    .silent = TRUE
  )

  normed <- update_norm_info(
    normed,
    list(
      .norm_procedure = "tidynorm::norm_track_lobanov"
    )
  )

  wrap_up(normed, .silent)

  return(normed)
}

#' Nearey Track Normalization
#'
#' @inheritParams norm_track_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#' When formant extrinsic:
#' \deqn{
#'  \hat{F}_{ij} = \log(F_{ij}) - L
#' }
#' \deqn{
#'  L = \frac{1}{MN}\sum_{i=1}^M\sum_{j=1}^N \log(F_{ij})
#' }
#'
#' When formant intrinsic:
#' \deqn{
#'  \hat{F}_{ij} = \log(F_{ij}) - L_{i}
#' }
#'
#' \deqn{
#'   L_i = \frac{1}{N}\sum_{j=1}^{N}\log(F_{ij})
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data frame of Nearey normalized formant tracks.
#'
#' @example inst/examples/ex-norm_track_nearey.R
#'
#' @references
#' Nearey, T. M. (1978). Phonetic Feature Systems for Vowels \[Ph.D.\].
#' University of Alberta.
#' @export
norm_track_nearey <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .by_formant = FALSE,
    .time_col = NULL,
    .order = 5,
    .return_dct = FALSE,
    .drop_orig = FALSE,
    .names = "{.formant}_lm",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- expr(...)
  normed <- norm_track_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .by_formant = .by_formant,
    .L = mean(!!sym(".formant"), na.rm = T),
    .S = (1 / sqrt(2)),
    .pre_trans = log,
    .post_trans = identity,
    .time_col = {{ .time_col }},
    .order = .order,
    .return_dct = .return_dct,
    .drop_orig = .drop_orig,
    .names = .names,
    .silent = TRUE
  )

  normed <- update_norm_info(
    normed,
    list(
      .norm_procedure = "tidynorm::norm_track_nearey"
    )
  )

  wrap_up(normed, .silent)

  return(normed)
}



#' Delta F Track Normalization
#'
#' @inheritParams norm_track_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#'
#' @details
#' \deqn{
#'  \hat{F}_{ij} = \frac{F_{ij}}{S}
#' }
#' \deqn{
#'  S = \frac{1}{MN}\sum_{i=1}^M\sum_{j=1}^N \frac{F_{ij}}{i-0.5}
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data frame of Delta F normalized formant tracks.
#'
#' @example inst/examples/ex-norm_track_deltaF.R
#'
#' @references
#' Johnson, K. (2020). The \eqn{\Delta}F method of vocal tract length normalization for vowels.
#' Laboratory Phonology: Journal of the Association for Laboratory Phonology, 11(1),
#' Article 1. \doi{10.5334/labphon.196}
#'
#' @export
norm_track_deltaF <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .time_col = NULL,
    .order = 5,
    .return_dct = FALSE,
    .drop_orig = FALSE,
    .names = "{.formant}_df",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- expr(...)
  normed <- norm_track_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .by_formant = FALSE,
    .L = 0,
    .S = mean(!!sym(".formant") / (!!sym(".formant_num") - 0.5), na.rm = T),
    .pre_trans = identity,
    .post_trans = identity,
    .time_col = {{ .time_col }},
    .order = .order,
    .return_dct = .return_dct,
    .drop_orig = .drop_orig,
    .names = .names,
    .silent = TRUE
  )
  normed <- update_norm_info(
    normed,
    list(
      .norm_procedure = "tidynorm::norm_track_deltaF"
    )
  )

  wrap_up(normed, .silent)

  return(normed)
}


#' Watt and Fabricius Track normalization
#'
#' @inheritParams norm_track_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#' This is a modified version of the Watt & Fabricius Method. The original
#' method identified point vowels over which F1 and F2 centroids were calculated.
#' The procedure here just identifies centroids by taking the mean of
#' all formant values.
#'
#' \deqn{
#' \hat{F}_{ij} = \frac{F_{ij}}{S_i}
#' }
#'
#' \deqn{
#'  S_i = \frac{1}{N}\sum_{j=1}^N F_{ij}
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data frame of Watt & Fabricius normalized formant tracks.
#'
#' @example inst/examples/ex-norm_track_wattfab.R
#'
#' @references
#' Watt, D., & Fabricius, A. (2002). Evaluation of a technique for improving the
#' mapping of multiple speakers’ vowel spaces in the F1 ~ F2 plane.
#' Leeds Working Papers in Linguistics and Phonetics, 9, 159–173.
#' @export
norm_track_wattfab <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .time_col = NULL,
    .order = 5,
    .return_dct = FALSE,
    .drop_orig = FALSE,
    .names = "{.formant}_wf",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- expr(...)
  normed <- norm_track_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .by_formant = TRUE,
    .L = 0,
    .S = mean(!!sym(".formant"), na.rm = T),
    .pre_trans = identity,
    .post_trans = identity,
    .time_col = {{ .time_col }},
    .order = .order,
    .return_dct = .return_dct,
    .drop_orig = .drop_orig,
    .names = .names,
    .silent = TRUE
  )

  normed <- update_norm_info(
    normed,
    list(
      .norm_procedure = "tidynorm::norm_track_wattfab"
    )
  )

  wrap_up(normed, .silent)

  return(normed)
}

#' Bark Difference Track Normalization
#' @inheritParams norm_track_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#' This is a within-token normalization technique. First all formants are
#' converted to Bark (see [hz_to_bark]), then, within each token, F3 is
#' subtracted from F1 and F2.
#'
#' \deqn{
#' \hat{F}_{ij} = F_{ij} - L_j
#' }
#'
#' \deqn{
#' L_j = F_{3j}
#' }
#'
#' @returns
#' A data frame of either normalized formant tracks, or normalized DCT
#' parameters.
#'
#' @returns
#' A data frame of Bark difference normalized formant tracks.
#'
#' @references
#' Syrdal, A. K., & Gopal, H. S. (1986). A perceptual model of vowel
#' recognition based on the auditory representation of American English vowels.
#' The Journal of the Acoustical Society of America, 79(4), 1086–1100.
#' \doi{10.1121/1.393381}
#'
#' @example inst/examples/ex-norm_track_barkz.R
#'
#' @export
norm_track_barkz <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .time_col = NULL,
    .order = 5,
    .return_dct = FALSE,
    .drop_orig = FALSE,
    .names = "{.formant}_bz",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- rlang::expr(c(...))
  target_pos <- tidyselect::eval_select(targets, .data)
  formant_nums <- name_to_formant_num(names(target_pos))

  if (length(target_pos) < 3) {
    cli_abort(
      message = c(
        "{.fn tidynorm::norm_track_barkz} requires F3."
      )
    )
  }


  f3 <- names(target_pos)[formant_nums == 3]

  targets <- expr(...)
  normed <- norm_track_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .by_formant = FALSE,
    .by_token = TRUE,
    .L = (!!sym(".formant"))[3],
    .S = (1 / sqrt(2)),
    .pre_trans = hz_to_bark,
    .post_trans = identity,
    .time_col = {{ .time_col }},
    .order = .order,
    .return_dct = .return_dct,
    .drop_orig = .drop_orig,
    .names = .names,
    .silent = TRUE
  )

  normed <- update_norm_info(
    normed,
    list(
      .norm_procedure = "tidynorm::norm_track_barkz",
      .f3 = f3
    )
  )

  wrap_up(normed, .silent)

  return(normed)
}

norm_messages <- list(
  .step = "{ .step}",
  .norm_procedure = "normalized with {.fn { .norm_procedure}}",
  .targets = "normalized {.var { .targets}}",
  .f3  = "{.var { .f3}} used for third formant.",
  .norm_cols = "normalized values in {.var { .norm_cols}}",
  .token_id_col = "token id column: {.var { .token_id_col}}",
  .time_col = "time column: {.var { .time_col}}",
  .param_col = "DCT parameter column: {.var { .param_col}}",
  .by = "grouped by {.var { .by}}",
  .by_formant = "within formant: { .by_formant}",
  .by_token = "within token: { .by_token}",
  .pre_trans = "Transformation prior to normalization: { .pre_trans}",
  .norm = "{ .norm}",
  .post_trans = "Transformation after normalization: { .post_trans}"
)

number_names <- c(
  one = 1,
  two = 2,
  three = 3,
  four = 4,
  five = 5,
  first = 1,
  second = 2,
  third = 3,
  fourth = 4,
  fifth = 5
)

check_grouping <- function(
    .data,
    .by,
    call = caller_env()) {
  grouped_by <- dplyr::group_vars(.data)
  grouping <- try_fetch(
    tidyselect::eval_select(enquo(.by), .data),
    error = \(cnd) selection_errors(cnd, arg = ".by", call = call)
  )
  if (length(grouped_by) > 0 & length(grouping) > 0) {
    cli_abort(
      c(
        "You cannot provide {.arg .by} to a grouped data frame.",
        "i" = "{.fn dplyr::group_by} was used on {.var .data}, and it is still grouped.",
        "i" = "You may want to {.fn dplyr::ungroup} {.var .data} before normalizing."
      ),
      call = call
    )
  }

  if (length(grouped_by) < 1 & length(grouping) < 1 & options::opt("tidynorm.warnings")) {
    cli_par()
    cli_warn(
      c(
        "There is was no grouping provided.",
        "i" = "You may want to provide {.arg .by} with a speaker id column."
      ),
      call = call
    )
    cli_end()
  }
}

check_tokens <- function(
    tokens,
    call = caller_env()) {
  if (length(tokens) < 1) {
    cli_abort(
      c(
        "No column passed to {.arg .token_id_col}.",
        "i" = "Provide column name(s) that uniquely identify tokens."
      )
    )
  }
}

check_args <- function(
    args,
    fmls,
    call = caller_env()) {
  fmls_undot <- stringr::str_remove(
    fmls,
    "^\\."
  )
  args <- args[nzchar(args)]
  if (any(!args %in% fmls)) {
    offenders <- args[!args %in% fmls]
    indef <- ""
    message <- c(
      "{.arg {offenders}} {?is/are} not{? a / }valid argument{?s}"
    )
    if (any(args %in% fmls_undot)) {
      undotted <- args[args %in% fmls_undot]
      redotted <- stringr::str_c(".", undotted)
      message <- c(
        message,
        "i" = "Should {.arg {undotted}} be {.arg {redotted}}?"
      )
    }
    cli_abort(
      message = message,
      call = call
    )
  }
}

selection_errors <- function(
    cnd,
    arg = "",
    call = caller_env()) {
  if (cnd_inherits(cnd, "vctrs_error_subscript_oob")) {
    cli_abort(
      "Problem with column selection for {.arg {arg}}",
      parent = cnd,
      call = call
    )
  } else if (cnd_inherits(cnd, "vctrs_error_subscript")) {
    cli_abort(
      c(
        "Problem with column selection for {.arg {arg}}",
        "i" = "Most arguments need to start with a .",
        "i" = "{.arg .{arg}}"
      ),
      parent = cnd,
      call = call
    )
  } else {
    cli_abort(
      "Problem with column selection for {.arg {arg}}",
      parent = cnd,
      call = call
    )
  }
}

make_dct_grouping <- function(
    .data,
    .by,
    .token_id_col) {
  cols <- enquos(
    .by = .by,
    .token_id_col = .token_id_col
  )
  if (length(dplyr::group_vars(.data)) > 0) {
    by_grouping <- expr(NULL)
    .data <- dplyr::group_by(
      .data,
      {{ .token_id_col }},
      .add = TRUE
    )
    joining <- dplyr::group_vars(.data)
  } else {
    by_grouping <- expr(c({{ .by }}, {{ .token_id_col }}))
    joining <- c()
    for (col in cols) {
      joining <- c(
        joining,
        names(
          tidyselect::eval_select(col, .data)
        )
      )
    }
  }

  return(
    list(
      .data = .data,
      by_grouping = by_grouping,
      joining = joining
    )
  )
}

append_norm_info <- function(
    .data,
    info) {
  prev_attr <- attributes(.data)$norminfo

  if (length(info$.post_trans) > 1) {
    info$.post_trans <- info$.post_trans[3]
  }

  if (length(info$.pre_trans) > 1){
    info$.pre_trans <- info$.pre_trans[3]
  }

  if ("identity" %in% info$.post_trans) {
    info$.post_trans <- NULL
  }

  if ("identity" %in% info$.pre_trans) {
    info$.pre_trans <- NULL
  }

  attr(.data, "norminfo") <- c(
    prev_attr,
    list(
      info
    )
  )

  .data
}

update_norm_info <- function(
    .data,
    info) {
  norminfo <- attr(.data, "norminfo")
  last_norm <- norminfo[[length(norminfo)]]

  last_names <- names(last_norm)
  new_names <- names(info)

  for (n in new_names) {
    last_norm[[n]] <- info[[n]]
  }

  norminfo[[length(norminfo)]] <- last_norm
  attr(.data, "norminfo") <- norminfo

  .data
}

wrap_up <- function(.data, .silent = FALSE) {

  if(.silent) return()


  if (is.null(attr(.data, "norminfo"))) {
    cli_par()
    cli_inform(
      "Not normalized with a {.pkg tidynorm} procedure."
    )
    return()
    cli_end()
  }

  norminfo <- attr(.data, "norminfo")
  last_norm <- norminfo[[length(norminfo)]]

  message <- c("Normalization info")

  for (n in names(norm_messages)) {
    if (n %in% names(last_norm)) {
      message <- c(
        message,
        "*" = norm_messages[[n]]
      )
    }
  }

  cli_par()
  cli_inform(
    message,
    .envir = last_norm
  )
  cli_end()
}

#' Check Normalization Procedures
#'
#' `check_norm()` will generate a message with information
#' about which normalization procedures have been applied to the
#' data.
#'
#' @param .data A data frame produced by a tidynorm function.
#'
#' @returns
#' This only prints an info message.
#'
#' @examples
#' speaker_norm <- speaker_data |>
#'   norm_nearey(
#'     F1:F3,
#'     .by = speaker,
#'     .silent = TRUE
#'   )
#'
#' check_norm(speaker_norm)
#'
#' @export
check_norm <- function(.data) {
  if (is.null(attr(.data, "norminfo"))) {
    cli_par()
    cli_inform(
      "x" = "Not normalized with a {.pkg tidynorm} procedure."
    )
    return()
    cli_end()

    return()
  }

  norminfo <- attr(.data, "norminfo")

  for (step in norminfo) {
    message <- "Normalization Step"
    for (n in names(norm_messages)) {
      if (n %in% names(step)) {
        message <- c(
          message,
          "*" = norm_messages[[n]]
        )
      }
    }
    cli_par()
    cli_inform(message, .envir = step)
    cli_end()
  }
}

#' Convert text-based formant names to
#' numeric
#' @importFrom checkmate check_character
#' @noRd
name_to_formant_num <- function(.formant_name, call = caller_env()) {
  checkmate::check_character(.formant_name, any.missing = FALSE, min.len = 1)

  has_digit <- stringr::str_detect(
     .formant_name, r"{[fF][1-5]}"
     ) |>
       all()

  if (has_digit) {
    nums = stringr::str_extract(
      .formant_name,
      r"{[fF]([1-5])}",
      group = 1
    ) |>
      as.numeric()
    return(nums)
  }

  has_name <- stringr::str_detect(
    .formant_name,
    stringr::regex(
      stringr::str_flatten(names(number_names), collapse = "|"),
      ignore_case = TRUE
    )
  ) |>
    all()

  if (has_name) {
    digit_name <- stringr::str_extract(
      .formant_name,
      stringr::regex(
        stringr::str_flatten(names(number_names), collapse = "|"),
        ignore_case = TRUE
      )
    ) |>
      tolower()
    nums = number_names[digit_name] |>
      unname()

    return(nums)
  }

  as.numeric(as.factor(.formant_name))

}

#' Hz to Bark
#'
#' Converts Hz to Bark
#'
#' @param hz Frequency in Hz
#'
#' @details
#' \deqn{
#' \hat{b} = \frac{26.81 hz}{1960 + hz} - 0.53
#' }
#' \deqn{
#' b = \begin{cases}
#' \hat{b} + 0.15(2-\hat{b}) & \text{if}~\hat{b} < 2\\
#' \hat{b} + 0.22(\hat{b} - 20.1) & \text{if}~\hat{b} > 20.1\\
#' \hat{b} & \text{otherwise}
#' \end{cases}
#' }
#'
#' @returns
#' A vector of bark scaled values
#'
#' @references
#' Traunmüller, H. (1990). Analytical expressions for the tonotopic
#' sensory scale. The Journal of the Acoustical Society of America,
#' 88(1), 97–100. \doi{10.1121/1.399849}
#'
#' @examples
#' hz <- seq(150, 2000, length = 100)
#' bark <- hz_to_bark(hz)
#' plot(hz, bark)
#'
#' @export
hz_to_bark <- function(hz) {
  bark <- ((26.81 * hz) / (1960 + hz)) - 0.53

  bark <- dplyr::case_when(
    bark < 2 ~ bark + ((0.15) * (2 - bark)),
    bark > 20.1 ~ bark + (0.22 * (bark - 20.1)),
    .default = bark
  )

  return(bark)
}

#' Bark to Hz
#'
#' Converts bark to Hz
#'
#' @param bark Frequency in Bark
#' @details
#' \deqn{
#' \hat{b} = \begin{cases}
#' \frac{b - 0.3}{0.85} & \text{if} ~ b < 2\\
#' \frac{b + 4.422}{1.22} & \text{if} ~ b > 20.1\\
#' b & \text{otherwise}
#' \end{cases}
#' }
#'
#' \deqn{
#' hz = 1960\frac{\hat{b} + 0.53}{26.28 - \hat{b}}
#' }
#'
#' @returns
#' A vector of Hz scaled values
#'
#' @references
#' Traunmüller, H. (1990). Analytical expressions for the tonotopic
#' sensory scale. The Journal of the Acoustical Society of America,
#' 88(1), 97–100. \doi{10.1121/1.399849}
#'
#' @examples
#' bark <- seq(1.5, 13, length = 100)
#' hz <- bark_to_hz(bark)
#' plot(bark, hz)
#' @export
bark_to_hz <- function(bark) {
  bark <- dplyr::case_when(
    bark < 2 ~ (bark - 0.3) / 0.85,
    bark > 20.1 ~ (bark + 4.422) / 1.22,
    .default = bark
  )
  hz <- 1960 * ((bark + 0.53) / (25.28 - bark))
  return(hz)
}


#' Hz to Mel
#'
#' Convert Hz to Mel
#'
#' @param hz Numeric values in Hz
#' @param htk Whether or not to use the HTK formula
#'
#' @details
#' This is a direct re-implementation of the `hz_to_mel`
#' function from the [librosa](https://librosa.org/) library.
#'
#' The default method is to use the method due to Slaney (1998), which
#' is linear below 1000Hz, and logarithmic above.
#'
#' If `htk=TRUE`, the method from HTK, due to O'Shaughnessy (1987) is used.
#'
#' @returns A numeric vector of Mel values
#'
#' @references
#' McFee, B., C. Raffel, D. Liang, D. PW Ellis, M. McVicar, E. Battenberg,
#' and O. Nieto. librosa: Audio and music signal analysis in python.
#' In Proceedings of the 14th python in science conference, pp. 18-25.
#'
#' O'Shaughnessy, D (1987). Speech communication: human and machine.
#' Addison-Wesley. p. 150. ISBN 978-0-201-16520-3.
#'
#' Slaney, M. (1998) Auditory Toolbox: A MATLAB Toolbox for Auditory Modeling
#' Work. Technical Report, version 2, Interval Research Corporation.
#'
#' @examples
#' hz_to_mel(c(500, 1000, 2000, 3000))
#'
#' @export
hz_to_mel <- function(
    hz,
    htk = FALSE) {
  if (htk) {
    mels <- 2595.0 * log10(1.0 + (hz / 700.0))
    return(mels)
  }
  f_sp <- 200 / 3
  min_log_hz <- 1000.0
  min_log_mel <- min_log_hz / f_sp
  logstep <- log(6.4) / 27.0

  mels <- dplyr::case_when(
    hz >= 1000 ~ min_log_mel + (log(hz / min_log_hz) / logstep),
    .default = hz / f_sp
  )

  return(mels)
}

#' Mel to Hz
#'
#' Convert Mel to Hz
#'
#' @param mel Numeric values in Hz
#' @param htk Whether or not to use the HTK formula
#'
#' @details
#' This is a direct re-implementation of the `hz_to_mel`
#' function from the [librosa](https://librosa.org/) library.
#'
#' The default method is to use the method due to Slaney (1998), which
#' is linear below 1000Hz, and logarithmic above.
#'
#' If `htk=TRUE`, the method from HTK, due to O'Shaughnessy (1987) is used.
#'
#' @returns A numeric vector of Hz values
#'
#' @references
#' McFee, B., C. Raffel, D. Liang, D. PW Ellis, M. McVicar, E. Battenberg,
#' and O. Nieto. librosa: Audio and music signal analysis in python.
#' In Proceedings of the 14th python in science conference, pp. 18-25.
#'
#' O'Shaughnessy, D (1987). Speech communication: human and machine.
#' Addison-Wesley. p. 150. ISBN 978-0-201-16520-3.
#'
#' Slaney, M. (1998) Auditory Toolbox: A MATLAB Toolbox for Auditory Modeling
#' Work. Technical Report, version 2, Interval Research Corporation.
#'
#'
#' @examples
#' mel_to_hz(c(7.5, 15, 25, 31))
#'
#' @export
mel_to_hz <- function(
    mel,
    htk = FALSE) {
  if (htk) {
    hz <- 700.0 * (10.0^(mel / 2595.0) - 1.0)
    return(hz)
  }

  f_sp <- 200.0 / 3
  min_log_hz <- 1000.0
  min_log_mel <- min_log_hz / f_sp
  logstep <- log(6.4) / 27.0

  hz <- dplyr::case_when(
    mel >= min_log_mel ~ min_log_hz * exp(logstep * (mel - min_log_mel)),
    .default = f_sp * mel
  )

  return(hz)
}

#' Speaker Data
#' @format ## `speaker_data`
#' A data frame with 10,697 rows and 8 columns
#' \describe{
#'   \item{speaker}{Speaker ID column}
#'   \item{vowel}{CMU Dictionary vowel class}
#'   \item{plt_vclass}{Modified Labov-Trager vowel class}
#'   \item{ipa_vclas}{IPA-like vowel class}
#'   \item{word}{Word that the vowel appeared in}
#'   \item{F1, F2, F3}{The first, second and third formants, in Hz}
#' }
"speaker_data"

#' Speaker Tracks
#' @format ## `speaker_tracks`
#' A data frame with 20,000 rows and 9 columns
#' \describe{
#'   \item{speaker}{Speaker ID column}
#'   \item{id}{Within speaker id for each token}
#'   \item{vowel}{CMU Dictionary vowel class}
#'   \item{plt_vclass}{Modified Labov-Trager vowel class}
#'   \item{ipa_vclas}{IPA-like vowel class}
#'   \item{word}{Word that the vowel appeared in}
#'   \item{t}{Measurement time point}
#'   \item{F1, F2, F3}{The first, second and third formants, in Hz}
#' }
"speaker_tracks"

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

cos_fun <- function(j, k, N) {
    .Call(`_tidynorm_cos_fun`, j, k, N)
}

cos_bank <- function(jj, kk) {
    .Call(`_tidynorm_cos_bank`, jj, kk)
}

dct_fun <- function(y, kk) {
    .Call(`_tidynorm_dct_fun`, y, kk)
}

dct_mat <- function(y, kk) {
    .Call(`_tidynorm_dct_mat`, y, kk)
}

idct_fun <- function(y, n) {
    .Call(`_tidynorm_idct_fun`, y, n)
}

idct_mat <- function(y, n) {
    .Call(`_tidynorm_idct_mat`, y, n)
}

idct_prime <- function(y, n) {
    .Call(`_tidynorm_idct_prime`, y, n)
}

idct_dprime <- function(y, n) {
    .Call(`_tidynorm_idct_dprime`, y, n)
}

idct_interp <- function(y, n, p) {
    .Call(`_tidynorm_idct_interp`, y, n, p)
}

#' Generic Formant DCT Normalization Procedure
#'
#' @inheritParams norm_track_generic
#' @param .data A data frame of formant DCT coefficients
#' @param .param_col A column identifying the DCT parameter number.
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#' @details
#'
#' The following `norm_dct_*` procedures were built on top of
#' `norm_dct_generic()`.
#'
#' - [norm_dct_lobanov]
#' - [norm_dct_nearey]
#' - [norm_dct_deltaF]
#' - [norm_dct_wattfab]
#' - [norm_dct_barkz]
#'
#' ## Normalizing DCT Coefficients
#'
#' This will normalize vowel formant data that has already had the
#' Discrete Cosine Transform applied (see [dct]) with the following
#' procedure:
#'
#' 1. Location `.L` and Scale `.S` expressions will be used to summarize
#' the zero<sup>th</sup> DCT coefficients.
#' 2. These location and scale will be used to normalize the DCT coefficients.
#'
#' ### Location and Scale expressions
#' [norm_dct_generic] normalizes DCT coefficients directly.
#' If \eqn{F_k} is the k<sup>th</sup> DCT coefficient
#' the normalization procedure is
#'
#' \deqn{
#' \hat{F}_k = \frac{F_k - L'}{\sqrt{2}S}
#' }
#' \deqn{
#' L' = \begin{cases}
#'    L & \text{for }k=0\\
#'    0 & \text{for }k>0
#'  \end{cases}
#' }
#'
#' Rather than requiring users to remember to multiply expressions for \eqn{S}
#' by \eqn{\sqrt{2}}, this is done by [norm_dct_generic] itself, to allow greater
#' parallelism with how [norm_generic] works.
#'
#' **Note**: If you want to scale values by a constant in the normalization,
#' you'll need to divide the constant by `sqrt(2)`.
#'
#' The expressions for calculating \eqn{L} and \eqn{S} can be
#' passed to `.L` and `.S`, respectively. Available values for
#' these expressions are
#'
#' \describe{
#'  \item{`.formant`}{The original formant value}
#'  \item{`.formant_num`}{The number of the formant. (e.g. 1 for F1, 2 for F2 etc)}
#' }
#'
#' Along with any data columns from your original data.
#'
#' ### Identifying tokens
#' DCT normalization requires identifying individual tokens, so there must be a column that
#' uniquely identifies (or, in combination with a `.by` grouping, uniquely
#' identifies) each individual token. This column should be passed to
#' `.token_id_col`.
#'
#' @returns
#' A data frame of normalized DCT coefficients.
#'
#' @example inst/examples/ex-norm_dct_generic.R
#'
#' @export
norm_dct_generic <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .param_col = NULL,
    .L = 0,
    .S = 1 / sqrt(2),
    .by_formant = FALSE,
    .by_token = FALSE,
    .names = "{.formant}_n",
    .silent = opt("tidynorm.silent"),
    .drop_orig = FALSE,
    .call = caller_env()) {
  if (env_name(.call) == "global") {
    .call <- current_env()
  }
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls, .call)

  prev_attr <- attributes(.data)$norminfo

  .names2 <- glue::glue(.names, .formant = ".formant")

  targets <- expr(c(...))

  target_pos <- try_fetch(
    tidyselect::eval_select(targets, data = .data),
    error = \(cnd) selection_errors(cnd, arg = "...", call = .call)
  )

  group_pos <- tidyselect::eval_select(
    enquo(.by),
    data = .data
  )

  token_pos <- tidyselect::eval_select(
    enquo(.token_id_col),
    data = .data
  )

  if (!".param_col" %in% args) {
    .param_col <- sym(".param")
  }

  cols <- enquos(
    .by = .by,
    .token_id_col = .token_id_col,
    .param_col = .param_col
  )

  for (x in cols) {
    try_fetch(
      tidyselect::eval_select(x, data = .data),
      error = \(cnd) selection_errors(cnd)
    )
  }

  check_grouping(.data, {{ .by }}, .call)

  grouping_list <- make_dct_grouping(
    .data,
    {{ .by }},
    {{ .token_id_col }}
  )
  .data <- grouping_list$.data
  by_grouping <- grouping_list$by_grouping
  joining <- grouping_list$joining

  .dct_data <- .data |>
    tidyr::pivot_longer(
      cols = !!targets,
      names_to = ".formant_name",
      values_to = ".formant"
    ) |>
    dplyr::mutate(
      .formant_num = stringr::str_extract(
        !!sym(".formant_name"),
        r"{\d}"
      ) |>
        as.numeric()
    )

  grouped_by <- length(dplyr::group_vars(.dct_data)) > 0

  # augment grouping as necessary to
  # match .by_formant
  # byformant <- NULL
  if (.by_formant & grouped_by) {
    .dct_data <- dplyr::group_by(
      .dct_data,
      !!sym(".formant_name"),
      .add = TRUE
    )
    by_grouping2 <- expr(NULL)
    by_grouping_noid <- expr(NULL)
  } else if (.by_token & grouped_by) {
    .dct_data <- dplyr::group_by(
      .dct_data,
      {{ .token_id_col }},
      .add = TRUE
    )
    by_grouping2 <- expr(NULL)
    by_grouping_noid <- expr(NULL)
  } else if (grouped_by) {
    by_grouping2 <- expr(NULL)
    by_grouping_noid <- expr(NULL)
  } else if (.by_formant) {
    by_grouping2 <- expr(c(!!by_grouping, sym(".formant_name")))
    by_grouping_noid <- expr(c({{ .by }}, sym(".formant_name")))
  } else if (.by_token) {
    by_grouping2 <- expr(c(!!by_grouping, {{ .token_id_col }}))
    by_grouping_noid <- expr(c({{ .by }}, {{ .token_id_col }}))
  } else {
    by_grouping2 <- by_grouping
    by_grouping_noid <- expr({{ .by }})
  }

  zeroth <- dplyr::filter(
    .dct_data,
    .by = !!by_grouping2,
    {{ .param_col }} == min({{ .param_col }})
  )

  if (grouped_by & !.by_token) {
    zeroth <- zeroth |>
      dplyr::ungroup({{ .token_id_col }})
  }

  if (grouped_by) {
    keepit <- "keep"
  } else {
    keepit <- NULL
  }

  norm_params <- zeroth |>
    dplyr::summarise(
      .by = !!by_grouping_noid,
      .L = {{ .L }},
      .S = {{ .S }} * sqrt(2),
      .groups = keepit
    )

  param_joining <- c(
    dplyr::group_vars(norm_params),
    names(
      tidyselect::eval_select(by_grouping_noid, norm_params)
    )
  )

  .dct_with_norm <- dplyr::left_join(
    .dct_data,
    norm_params,
    by = param_joining
  )

  normed <- .dct_with_norm |>
    dplyr::mutate(
      "{.names2}" := case_when(
        {{ .param_col }} == 0 ~ (!!sym(".formant") - !!sym(".L")) / !!sym(".S"),
        .default = !!sym(".formant") / !!sym(".S")
      )
    ) |>
    dplyr::select(
      -c(
        !!sym(".L"),
        !!sym(".S"),
        !!sym(".formant_num")
      )
    )

  if (.drop_orig) {
    normed <- dplyr::select(
      normed,
      -!!sym(".formant")
    )
  }

  normed <- normed |>
    tidyr::pivot_wider(
      names_from = !!sym(".formant_name"),
      values_from = c(starts_with(".formant") & !any_of(".formant_name")),
      names_glue = "{.formant_name}_{.value}"
    )

  call_tree <- trace_back()$call |>
    purrr::map(as_label) |>
    unlist()

  if (!"norm_track_generic(...)" %in% call_tree) {
    normed <- normed |>
      dplyr::rename_with(
        .fn = \(x) stringr::str_remove(x, "_.formant")
      )
  }

  attr(normed, "norminfo") <- prev_attr
  attr(normed, "normalized") <- TRUE
  norm_info <- list(
    .norm_procedure = "tidynorm::norm_dct_generic",
    .by_col = .by_formant,
    .targets = names(target_pos),
    .norm_cols = glue::glue(.names, .formant = names(target_pos)),
    .by = names(group_pos),
    .token_id_col = quo_name(enquo(.token_id_col)),
    .param_col = quo_name(enquo(.param_col)),
    .by_formant = .by_formant,
    .norm = glue::glue("(.formant - {quo_name(enquo(.L))})/{quo_name(enquo(.S))}")
  )

  if (.by_token) {
    norm_info <- c(
      norm_info,
      list(.by_token = .by_token)
    )
  }

  normed <- append_norm_info(
    normed,
    norm_info
  )

  wrap_up(normed, .silent)

  return(normed)
}

#' Lobanov DCT Normalization
#'
#' @inheritParams norm_dct_generic
#' @inheritParams norm_track_lobanov
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#'
#' \deqn{
#'   \hat{F}_{ij} = \frac{F_{ij} - L_i}{S_i}
#' }
#'
#' \deqn{
#'   L_i = \frac{1}{N}\sum_{j=1}^{N}F_{ij}
#' }
#'
#' \deqn{
#'   S_i = \sqrt{\frac{\sum(F_{ij}-L_i)^2}{N-1}}
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data frame of Lobanov normalized DCT Coefficients.
#'
#' @references
#' Lobanov, B. (1971). Classification of Russian vowels spoken by different listeners.
#' Journal of the Acoustical Society of America, 49, 606–608.
#'
#' @example inst/examples/ex-norm_dct_lobanov.R
#' @export
norm_dct_lobanov <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .param_col = NULL,
    .names = "{.formant}_z",
    .silent = opt("tidynorm.silent"),
    .drop_orig = FALSE) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- expr(...)
  normed <- norm_dct_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .by_formant = TRUE,
    .by_token = FALSE,
    .L = mean(!!sym(".formant"), na.rm = T),
    .S = sd(!!sym(".formant"), na.rm = T),
    .param_col = {{ .param_col }},
    .drop_orig = .drop_orig,
    .names = .names,
    .silent = TRUE
  )

  normed <- update_norm_info(
    normed,
    list(.norm_procedure = "tidynorm::norm_dct_lobanov")
  )

  wrap_up(normed, .silent)

  return(normed)
}


#' Nearey DCT Normalization
#'
#'
#' @inheritParams norm_track_generic
#' @inheritParams norm_dct_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#'
#' **Important**: This function assumes that the DCT
#' coefficients were estimated over log-transformed
#' formant values.
#'
#' When formant extrinsic:
#' \deqn{
#'  \hat{F}_{ij} = \log(F_{ij}) - L
#' }
#' \deqn{
#'  L = \frac{1}{MN}\sum_{i=1}^M\sum_{j=1}^N \log(F_{ij})
#' }
#'
#' When formant intrinsic:
#' \deqn{
#'  \hat{F}_{ij} = \log(F_{ij}) - L_{i}
#' }
#'
#' \deqn{
#'   L_i = \frac{1}{N}\sum_{j=1}^{N}\log(F_{ij})
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data frame of Nearey normalized DCT coefficients
#'
#' @example inst/examples/ex-norm_dct_nearey.R
#'
#' @references
#' Nearey, T. M. (1978). Phonetic Feature Systems for Vowels \[Ph.D.\].
#' University of Alberta.
#' @export
norm_dct_nearey <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .by_formant = FALSE,
    .param_col = NULL,
    .drop_orig = FALSE,
    .names = "{.formant}_lm",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- expr(...)
  normed <- norm_dct_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .by_formant = .by_formant,
    .by_token = FALSE,
    .L = mean(!!sym(".formant"), na.rm = T),
    .S = (1 / sqrt(2)),
    .param_col = {{ .param_col }},
    .drop_orig = .drop_orig,
    .names = .names,
    .silent = TRUE
  )

  normed <- update_norm_info(
    normed,
    list(
      .norm_procedure = "tidynorm::norm_dct_nearey"
    )
  )

  wrap_up(normed, .silent)

  return(normed)
}


#' Delta F DCT Normalization
#'
#' @inheritParams norm_track_generic
#' @inheritParams norm_dct_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#'
#' @details
#' \deqn{
#'  \hat{F}_{ij} = \frac{F_{ij}}{S}
#' }
#' \deqn{
#'  S = \frac{1}{MN}\sum_{i=1}^M\sum_{j=1}^N \frac{F_{ij}}{i-0.5}
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data frame of Delta F normalized DCT coefficients.
#'
#' @example inst/examples/ex-norm_dct_deltaF.R
#'
#' @references
#' Johnson, K. (2020). The \eqn{\Delta}F method of vocal tract length normalization for vowels.
#' Laboratory Phonology: Journal of the Association for Laboratory Phonology, 11(1),
#' Article 1. \doi{10.5334/labphon.196}
#'
#' @export
norm_dct_deltaF <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .param_col = NULL,
    .drop_orig = FALSE,
    .names = "{.formant}_df",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- expr(...)
  normed <- norm_dct_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .by_formant = FALSE,
    .L = 0,
    .S = mean(!!sym(".formant") / (!!sym(".formant_num") - 0.5), na.rm = T),
    .param_col = {{ .param_col }},
    .drop_orig = .drop_orig,
    .names = .names,
    .silent = TRUE
  )
  normed <- update_norm_info(
    normed,
    list(
      .norm_procedure = "tidynorm::norm_dct_deltaF"
    )
  )

  wrap_up(normed, .silent)

  return(normed)
}

#' Watt and Fabricius DCT normalization
#'
#' @inheritParams norm_track_generic
#' @inheritParams norm_dct_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#'
#' @details
#' This is a modified version of the Watt & Fabricius Method. The original
#' method identified point vowels over which F1 and F2 centroids were calculated.
#' The procedure here just identifies centroids by taking the mean of
#' all formant values.
#'
#' \deqn{
#' \hat{F}_{ij} = \frac{F_{ij}}{S_i}
#' }
#'
#' \deqn{
#'  S_i = \frac{1}{N}\sum_{j=1}^N F_{ij}
#' }
#'
#' Where
#' - \eqn{\hat{F}} is the normalized formant
#' - \eqn{i} is the formant number
#' - \eqn{j} is the token number
#'
#' @returns
#' A data frame of Watt & Fabricius normalized DCT coefficients.
#'
#' @example inst/examples/ex-norm_dct_wattfab.R
#'
#' @references
#' Watt, D., & Fabricius, A. (2002). Evaluation of a technique for improving the
#' mapping of multiple speakers’ vowel spaces in the F1 ~ F2 plane.
#' Leeds Working Papers in Linguistics and Phonetics, 9, 159–173.
#' @export
norm_dct_wattfab <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .param_col = NULL,
    .drop_orig = FALSE,
    .names = "{.formant}_wf",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- expr(...)
  normed <- norm_dct_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .by_formant = TRUE,
    .L = 0,
    .S = mean(!!sym(".formant"), na.rm = T),
    .param_col = {{ .param_col }},
    .drop_orig = .drop_orig,
    .names = .names,
    .silent = TRUE
  )

  normed <- update_norm_info(
    normed,
    list(
      .norm_procedure = "tidynorm::norm_dct_wattfab"
    )
  )

  wrap_up(normed, .silent)

  return(normed)
}

#' Bark Difference DCT Normalization
#' @inheritParams norm_track_generic
#' @inheritParams norm_dct_generic
#' @eval options::as_params(".silent" = "tidynorm.silent")
#' @importFrom options opt
#' @details
#'
#' **Important**: This function assumes that the DCT
#' coefficients were estimated over bark-transformed
#' formant values.
#'
#' This is a within-token normalization technique. First all formants are
#' converted to Bark (see [hz_to_bark]), then, within each token, F3 is
#' subtracted from F1 and F2.
#'
#' \deqn{
#' \hat{F}_{ij} = F_{ij} - L_j
#' }
#'
#' \deqn{
#' L_j = F_{3j}
#' }
#'
#' @returns
#' A data frame of normalized DCT
#' parameters.
#'
#' @returns
#' A data frame of Back Difference normalized dct coefficients.
#'
#' @references
#' Syrdal, A. K., & Gopal, H. S. (1986). A perceptual model of vowel
#' recognition based on the auditory representation of American English vowels.
#' The Journal of the Acoustical Society of America, 79(4), 1086–1100.
#' \doi{10.1121/1.393381}
#'
#' @example inst/examples/ex-norm_dct_barkz.R
#'
#' @export
norm_dct_barkz <- function(
    .data,
    ...,
    .token_id_col,
    .by = NULL,
    .param_col = NULL,
    .drop_orig = FALSE,
    .names = "{.formant}_bz",
    .silent = opt("tidynorm.silent")) {
  args <- names(call_match())
  fmls <- names(fn_fmls())
  check_args(args, fmls)

  targets <- rlang::expr(c(...))
  target_pos <- tidyselect::eval_select(targets, .data)
  formant_nums <- name_to_formant_num(names(target_pos))

  if (length(target_pos) < 3) {
    cli_abort(
      message = c(
        "{.fn tidynorm::norm_dct_barkz} requires F3."
      )
    )
  }

  f3 <- names(target_pos)[formant_nums == 3]

  normed <- norm_dct_generic(
    .data,
    !!targets,
    .by = {{ .by }},
    .token_id_col = {{ .token_id_col }},
    .by_formant = FALSE,
    .by_token = TRUE,
    .L = (!!sym(".formant"))[3],
    .S = (1 / sqrt(2)),
    .param_col = {{ .param_col }},
    .drop_orig = .drop_orig,
    .names = .names,
    .silent = TRUE
  )

  normed <- update_norm_info(
    normed,
    list(
      .norm_procedure = "tidynorm::norm_dct_barkz",
      .f3 = f3
    )
  )

  wrap_up(normed, .silent)

  return(normed)
}

library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

track_subset <- speaker_tracks |>
  filter(
    .by = c(speaker, id),
    if_all(
      F1:F3,
      .fns = \(x) mean(is.finite(x)) > 0.9
    ),
    row_number() %% 2 == 1
  )

track_norm <- track_subset |>
  norm_track_lobanov(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .drop_orig = TRUE
  )

if (ggplot2_inst) {
  track_norm |>
    ggplot(
      aes(F2_z, F1_z, color = speaker)
    ) +
    stat_density_2d(bins = 4) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}

# returning the DCT coefficients
track_norm_dct <- track_subset |>
  norm_track_lobanov(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .return_dct = TRUE,
    .drop_orig = TRUE,
    .names = "{.formant}_z"
  )

track_norm_means <- track_norm_dct |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      ends_with("_z"),
      mean
    )
  ) |>
  reframe_with_idct(
    ends_with("_z"),
    .by = speaker,
    .token_id_col = vowel,
    .param_col = .param
  )


if (ggplot2_inst) {
  track_norm_means |>
    ggplot(
      aes(F2_z, F1_z, color = speaker)
    ) +
    geom_path(
      aes(
        group = interaction(speaker, vowel)
      )
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}

---
title: "Checking Normalization Procedures"
knitr:
  opts_chunk:
    crop: !expr 'rlang::is_installed(c("magick"))'
    collapse: true
    comment: '#>'
format:
  html:
    toc: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{Checking Normalization Procedures}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| label: setup
library(tidynorm)
```

If you've carried out a few different normalization procedures on a single dataset, you can double check what the sequence of operations was with `check_norm()`.
For example, in this workflow, we've given the new normalized data columns uninformative names.

```{r}
norm_data <- speaker_data |>
  norm_lobanov(
    F1:F3,
    .by = speaker,
    .names = "{.formant}_norm1",
    .silent = T
  ) |>
  norm_nearey(
    F1:F3,
    .by = speaker,
    .names = "{.formant}_norm2",
    .silent = T
  )
```

We can review which normalization procedure produced which normalized column like so.

```{r}
check_norm(norm_data)
```

---
title: "Normalization Methods"
vignette: >
  %\VignetteIndexEntry{Normalization Methods}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
knitr:
  opts_chunk:
    crop: !expr 'rlang::is_installed(c("magick"))'
    collapse: true
    comment: '#>'
format:
  html:
    toc: true
    html-math-method: mathjax
bibliography: references.bib
---

```{r}
#| eval: !expr 'rlang::is_installed("magick")'
#| echo: false
knitr::knit_hooks$set(crop = knitr::hook_pdfcrop)
```

```{r}
#| label: setup
library(tidynorm)
library(dplyr)
```

In addition to the generic normalization functions in tidynorm (`norm_generic()`, `norm_track_generic()` and `norm_dct_generic()`), there are a number of convenience functions for a few established normalization methods.

## Lobanov [@lobanov]

tidynorm functions:

-   `norm_lobanov()`
-   `norm_track_lobanov()`
-   `norm_dct_lobanov()`

Lobanov normalization z-scores each formant.
If $F_{ij}$ is the $j^{th}$ token of the $i^{th}$ formant, and $\hat{F}_{ij}$ is its normalized value, then

$$
\hat{F}_{ij} = \frac{F_{ij} - L_i}{S_i}
$$

Where $L_i$ is the mean across the $i^{th}$ formant:

$$
L_i = \frac{1}{N}\sum_{j=1}^N F_{ij}
$$

And $S_i$ is the standard deviation across the $i^{th}$ formant.

$$
S_i = \sqrt{\frac{\sum_j(F_{ij}-L_i)^2}{N-1}}
$$

### Using the Lobanov normalization functions

#### On points

```{r}
point_norm <- speaker_data |>
  norm_lobanov(
    F1:F3,
    .by = speaker
  )
```

#### On tracks

```{r}
track_norm <- speaker_tracks |>
  norm_track_lobanov(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )
```

#### On DCT Coefficients

```{r}
dct_norm <- speaker_tracks |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  ) |>
  norm_dct_lobanov(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param
  )
```

## Nearey Normalization [@neareyPhoneticFeatureSystems1978]

tidynorm functions:

-   `norm_nearey()`
-   `norm_track_nearey()`
-   `norm_dct_nearey()`

Nearey Normalization first log transforms formant values, then subtracts the grand mean across all formants.
If $F_{ij}$ is the $j^{th}$ token of the $i^{th}$ formant, and $\hat{F}_{ij}$ is its normalized value, then

$$
\hat{F}_{ij} = \log(F_{ij}) - L
$$

$$
L = \frac{1}{MN}\sum_{i = 1}^M\sum_{j=1}^N \log(F_{ij})
$$

The fact that the grand mean is taken across all formants, it's important to report whether just F1 and F2 were used, or if F1, F2 and F3 were used.

### Using the Nearey normalization functions

#### On points

```{r}
point_norm <- speaker_data |>
  norm_nearey(
    F1:F3,
    .by = speaker
  )
```

#### On tracks

```{r}
track_norm <- speaker_tracks |>
  norm_track_nearey(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )
```

#### On DCT Coefficients

```{r}
dct_norm <- speaker_tracks |>
  mutate(across(F1:F3, log)) |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  ) |>
  norm_dct_nearey(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param
  )
```

## Delta F [@johnsonDFMethodVocal2020]

tidynorm functions:

-   `norm_deltaF()`

-   `norm_track_deltaF()`

-   `norm_dct_deltaF()`

The $\Delta F$ normalization method is based on the average of formant spacing.
If $F_{ij}$ is the $j^{th}$ token of the $i^{th}$ formant, and $\hat{F}_{ij}$ is its normalized value, then

$$
\hat{F} = \frac{F_{ij}}{S}
$$

$$
S = \frac{1}{MN} \sum_{i=1}^M\sum_{j=1}^N \frac{F_{ij}}{i-0.5}
$$

The fact that this method takes a weighted average across all formants, it's important to report whether just F1 and F2 were used, or if F1, F2 and F3 were used.

### Using the DeltaF normalization functions

#### On points

```{r}
point_norm <- speaker_data |>
  norm_deltaF(
    F1:F3,
    .by = speaker
  )
```

#### On tracks

```{r}
track_norm <- speaker_tracks |>
  norm_track_deltaF(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )
```

#### On DCT coefficients

```{r}
dct_norm <- speaker_tracks |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  ) |>
  norm_dct_deltaF(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param
  )
```

## Watt & Fabricious [@wattEvaluationTechniqueImproving2002]

tidynorm functions:

-   `norm_wattfab()`

-   `norm_track_wattfab()`

-   `norm_dct_wattfab()`

The Watt & Fabricious method attempt to center vowel spaces on their "center of gravity".
The original Watt & Fabricious method involved calculating average F1 and F2 values for point vowels.
In tidynorm, a modified version has been implemented that just uses the average over F1 and F2 as the centers of gravity.
If $F_{ij}$ is the $j^{th}$ token of the $i^{th}$ formant, and $\hat{F}_{ij}$ is its normalized value, then

$$
\hat{F_{ij}} = \frac{F_{ij}}{S_i}
$$

Where $S_i$ is the mean across the $i_{th}$ formant.

$$
S_i = \frac{1}{N} \sum_{j = 1}^N F_{ij}
$$

### Using the Watt & Fabricious normaliation functions

#### On points

```{r}
point_norm <- speaker_data |>
  norm_wattfab(
    F1:F3,
    .by = speaker
  )
```

#### On tracks

```{r}
track_norm <- speaker_tracks |>
  norm_track_wattfab(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )
```

#### On DCT coefficients

```{r}
dct_norm <- speaker_tracks |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  ) |>
  norm_dct_wattfab(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param
  )
```

## Bark Difference [@syrdalPerceptualModelVowel1986]

tidynorm functions

-   `norm_barkz()`

-   `norm_track_barkz()`

-   `norm_dct_barkz()`

The bark difference metric tries to normalize vowels on the basis of individual tokens.
First, formant data is converted to bark (see `hz_to_bark()`), then F3 is subtracted from F1 and F2.
If $F_{ij}$ is the $j^{th}$ token of the $i^{th}$ formant, and $\hat{F}_{ij}$ is its normalized value, then

$$
\hat{F}_{ij} = \text{bark}(F_{ij}) - L_j
$$

$$
L_j = \text{bark}(F_{3j})
$$

### Using the Bark Difference normalization functions

#### On points

```{r}
point_norm <- speaker_data |>
  norm_barkz(
    F1:F3,
    .by = speaker
  )
```

#### On tracks

```{r}
track_norm <- speaker_tracks |>
  norm_track_barkz(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )
```

#### On DCT Coefficients

```{r}
dct_norm <- speaker_tracks |>
  mutate(
    across(F1:F3, hz_to_bark)
  ) |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  ) |>
  norm_dct_barkz(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param
  )
```

## References

---
title: "Normalizing Formant Tracks"
fig-width: 5
fig-height: 3
out-width: 80%
knitr:
  opts_chunk:
    crop: !expr 'rlang::is_installed(c("magick"))'
    collapse: true
    comment: '#>'
format:
  html:
    toc: true
    html-math-method: mathjax
    embed-resources: true
    theme: none
    minimal: true
vignette: >
  %\VignetteIndexEntry{Normalizing Formant Tracks}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| eval: !expr 'rlang::is_installed("magick")'
#| echo: false
knitr::knit_hooks$set(crop = knitr::hook_pdfcrop)
```

```{r}
#| label: setup
library(tidynorm)
```

```{r}
#| eval: !expr 'rlang::is_installed(c("ggplot2"))'
library(ggplot2)
```

Tidynorm normalizes vowel formant tracks by

1.  Converting them to Discrete Cosine Transform coefficients.
2.  Directly normalizing the DCT coefficients.
3.  Applying the inverse DCT.

## Token IDs

In order to do this successfully, each vowel token needs to be uniquely identifiable with a token id, or a token id in combination with other columns.
For example, in `tidynorm::speaker_tracks` the column `speaker` combined with the column `id` identifies each unique vowel formant track.

```{r}
speaker_tracks
```

For most normalization procedures, we'll want to group the data by the speaker column anyway, and in those cases it's sufficient to pass `id` to the `.token_id_col` argument of `norm_track_*` functions.

We can Lobanov normalize these speakers' formant tracks with `norm_track_lobanov()`.

```{r}
normed_tracks <- speaker_tracks |>
  norm_track_lobanov(
    # identify the formant columns
    F1:F3,

    # provide speaker grouping
    .by = speaker,

    # provide token id
    .token_id_col = id,

    # provide an optional time column
    .time_col = t
  )
```

```{r}
#| fig-align: center
#| out-width: 80%
#| eval: !expr 'rlang::is_installed(c("ggplot2"))'
normed_tracks |>
  ggplot(
    aes(F2_z, F1_z)
  ) +
  geom_path(
    aes(
      group = interaction(speaker, id)
    ),
    alpha = 0.3
  ) +
  scale_y_reverse() +
  scale_x_reverse() +
  facet_wrap(~speaker) +
  coord_cartesian(
    xlim = c(3.5, -3.5),
    ylim = c(3.5, -3.5)
  ) +
  theme(
    aspect.ratio = 1
  )
```

---
title: "The Discrete Cosine Transform"
fig-width: 5
fig-height: 3
out-width: 80%
knitr:
  opts_chunk:
    crop: !expr 'rlang::is_installed(c("magick"))'
    collapse: true
    comment: '#>'
format:
  html:
    toc: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{The Discrete Cosine Transform}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| eval: !expr 'rlang::is_installed("magick")'
#| echo: false
knitr::knit_hooks$set(crop = knitr::hook_pdfcrop)
```

```{r}
#| label: setup
#| message: false
library(tidynorm)
library(dplyr)
library(tibble)
library(ggplot2)
```

```{r}
#| label: plotting defaults
#| code-fold: true
#| code-summary: plotting defaults

options(
  ggplot2.discrete.colour = c(
    lapply(
      1:6,
      \(x) c(
        "#4477AA", "#EE6677", "#228833",
        "#CCBB44", "#66CCEE", "#AA3377"
      )[1:x]
    )
  ),
  ggplot2.discrete.fill = c(
    lapply(
      1:6,
      \(x) c(
        "#4477AA", "#EE6677", "#228833",
        "#CCBB44", "#66CCEE", "#AA3377"
      )[1:x]
    )
  )
)

theme_set(
  theme_minimal(
    base_size = 16
  )
)
```

## A brief description

The Discrete Cosine Transform re-describes an input signal as a set of coefficients.
These coefficients can be converted back into the original signal, or simplified, to get back a smoothed form of the original signal.

For example here is an F1 track with 20 measurement points from the `speaker_tracks` data set.

```{r}
one_track <- speaker_tracks |>
  filter(
    speaker == "s01",
    id == 9
  )
```

```{r}
#| label: fig-f1
#| fig-cap: "F1 Formant Track"
one_track |>
  ggplot(aes(t, F1)) +
  geom_point() +
  geom_line()
```

If we apply `dct()` to the F1 track, we'll get back 20 DCT coefficients.

```{r}
dct(one_track$F1)
```

And, if we apply `idct()` to these coefficients, we'll get back the original track.

```{r}
#| label: fig-f1-idct
#| fig-cap: "idct(dct(F1))"
one_track |>
  mutate(
    F1_dct = dct(F1),
    F1_idct = idct(F1_dct)
  ) |>
  ggplot(
    aes(t, F1_idct)
  ) +
  geom_point() +
  geom_line()
```

However, if we apply `idct()` to just the first few DCT coefficients, we'll get back a smoothed version of the formant track.

```{r}
#| label: fig-f1-dctsmooth
#| fig-cap: DCT smoothed F1
one_track |>
  mutate(
    F1_dct = dct(F1),
    F1_idct = idct(F1_dct[1:5], n = n())
  ) |>
  ggplot(
    aes(t, F1_idct)
  ) +
  geom_point() +
  geom_line()
```

## DCT functions in tidynorm

There are three `reframe_with_*` functions in tidynorm.

-   `reframe_with_dct()`

    -   This will take a data frame of formant tracks, and return a data frame of DCT coefficients.

    -   You need to be able to identify which rows belong to individual tokens, and can identify a column for the time domain.

-   `reframe_with_idct()`

    -   This will take a data frame of DCT coefficients, and return a data frame of formant tracks.

    -   You need to be able to identify which rows belong to individual tokens, and can identify a column for the parameter number.

-   `reframe_with_dct_smooth()`

    -   This combines `reframe_with_dct()` and `reframe_with_idct()` into one step, taking in a data frame of formant tracks, and returning a data frame of smoothed formant tracks.

    -   You need to be able to identify which rows belong to individual tokens, and can identify a column for the time domain.

## Getting average formant tracks

To get average formant tracks for each vowel, you'll need to

1.  Reframe the original formant tracks as DCT coefficients.
2.  Average over each parameter number for every vowel.
3.  Reframe these averages using the inverse DCT.

```{r}
#| warning: false
# focusing on one speaker
one_speaker <- speaker_tracks |>
  filter(speaker == "s01")

dct_smooths <- one_speaker |>
  # step 1, reframing as dct coefficients
  reframe_with_dct(
    F1:F3,
    .token_id_col = id,
    .time_col = t
  ) |>
  # step 2, averaging over parameter number and vowel
  summarise(
    across(F1:F3, mean),
    .by = c(.param, plt_vclass)
  ) |>
  # step 3, reframing with inverse DCT
  reframe_with_idct(
    F1:F3,
    # this time, the id column is the vowel class
    .token_id_col = plt_vclass,
    .param_col = .param
  )
```

```{r}
#| label: fig-dct-smooths
#| fig-cap: DCT smoothed formant trajectories.
dct_smooths |>
  filter(
    plt_vclass %in% c("iy", "ey", "ay", "ay0", "oy")
  ) |>
  ggplot(
    aes(F2, F1)
  ) +
  geom_path(
    aes(
      group = plt_vclass,
      color = plt_vclass
    ),
    arrow = arrow()
  ) +
  scale_y_reverse() +
  scale_x_reverse()
```

## The DCT Basis

The DCT decomposes an input signal as a combination of weighted cosine functions, and returns those weights.
You can access the cosine functions it uses with `dct_basis()`.

```{r}
basis <- dct_basis(100, 5)
matplot(basis, type = "l", lty = 1, lwd = 2)
```

One way to think about it is that the DCT is using these cosine functions in a regression, and the values that get returned are the coefficients.

```{r}
dct(one_track$F1)[1:5]
```

```{r}
lm(
  one_track$F1 ~ dct_basis(20, 5) - 1
) |>
  coef()
```

For more details on the mathematical formulation of the DCT, see the `dct()` help page.

---
title: Normalization Overview
fig-width: 5
fig-height: 3
out-width: 80%
knitr:
  opts_chunk:
    crop: !expr 'rlang::is_installed(c("magick"))'
    collapse: true
    comment: '#>'
format:
  html:
    toc: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{Normalization Overview}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
---

```{r}
#| eval: !expr 'rlang::is_installed("magick")'
#| echo: false
knitr::knit_hooks$set(crop = knitr::hook_pdfcrop)
```

# Introduction to the problem

```{r}
#| message: false
library(tidynorm)
library(dplyr)
library(tidyr)
```

```{r}
#| eval: !expr 'rlang::is_installed(c("ggplot2", "ggdensity"))'
library(ggplot2)
library(ggdensity)
```

```{r}
#| eval: !expr 'rlang::is_installed(c("ggplot2", "ggdensity"))'
#| include: !expr 'rlang::is_installed(c("ggplot2", "ggdensity"))'
#| code-fold: true
#| code-summary: color palette
options(
  ggplot2.discrete.colour = c(
    lapply(
      1:6,
      \(x) c(
        "#4477AA", "#EE6677", "#228833",
        "#CCBB44", "#66CCEE", "#AA3377"
      )[1:x]
    )
  ),
  ggplot2.discrete.fill = c(
    lapply(
      1:6,
      \(x) c(
        "#4477AA", "#EE6677", "#228833",
        "#CCBB44", "#66CCEE", "#AA3377"
      )[1:x]
    )
  )
)

theme_set(
  theme_minimal(base_size = 16) +
    theme(
      panel.grid = element_blank()
    )
)
```

Vowel formant frequencies are heavily influenced by speakers' vocal tract length, such that equivalent vowels can have dramatically different formant frequencies

```{r}
speaker_data |>
  summarise(
    .by = c(speaker, vowel),
    across(
      F1:F3,
      \(x)mean(x, na.rm = T)
    )
  ) ->
speaker_means
```

```{r}
#| code-fold: true
#| code-summary: Plotting Code
#| eval: !expr 'rlang::is_installed(c("ggplot2", "ggdensity"))'
#| fig-align: center
#| out-width: 80%
speaker_means |>
  filter(
    vowel %in% c("IY", "UW", "AE", "AA")
  ) |>
  ggplot(
    aes(F2, F1)
  ) +
  geom_label(
    aes(
      label = vowel,
      fill = speaker
    ),
    color = "white"
  ) +
  scale_y_reverse() +
  scale_x_reverse()
```

```{r}
#| code-fold: true
#| code-summary: Vowel Comparison
speaker_means |>
  filter(
    vowel %in% c("IY", "UW", "AE", "AA")
  ) |>
  pivot_longer(
    F1:F3,
    names_to = ".formant_name",
    values_to = ".formant"
  ) |>
  pivot_wider(
    names_from = vowel,
    values_from = .formant
  ) |>
  arrange(
    .formant_name
  )
```

When conducting an analysis that combines multiple different speakers' data together, we often want to factor out this vocal tract length difference, so that we can instead focus our attention on the *relative* values of these vowel frequencies within speakers.

## An example

Speakers s01 and s03 in the `speakers_data` data set have very different vowel spaces that differ both in the location of their center points, and scaling.

```{r}
#| eval: !expr 'rlang::is_installed(c("ggplot2", "ggdensity"))'
#| fig-align: center
#| out-width: 80%
#| code-fold: true
#| code-summary: Plotting code
speaker_centers <- speaker_data |>
  summarise(
    .by = speaker,
    across(
      F1:F3,
      ~ mean(.x, na.rm = T)
    )
  )

speaker_data |>
  ggplot(
    aes(F2, F1, color = speaker)
  ) +
  stat_hdr(
    probs = c(0.95, 0.8, 0.5),
    fill = NA,
    alpha = 1,
    linewidth = 1
  ) +
  geom_point(
    data = speaker_centers,
    size = 5
  ) +
  scale_x_reverse() +
  scale_y_reverse()
```

One approach to moving speakers center points to a common location is to subtract the mean from both F1 and F2.

```{r}
speaker_data_centered <- speaker_data |>
  mutate(
    .by = speaker,
    across(
      F1:F3,
      ~ .x - mean(.x, na.rm = T)
    )
  )
```

Now s01 and s03's center points have a common location, but the scale of s03's vowels is still smaller than s01's.

```{r}
#| eval: !expr 'rlang::is_installed(c("ggplot2", "ggdensity"))'
#| fig-align: center
#| out-width: 80%
#| code-fold: true
#| code-summary: Plotting code
speaker_data_centered |>
  ggplot(
    aes(F2, F1, color = speaker)
  ) +
  stat_hdr(
    probs = c(0.95, 0.8, 0.5),
    fill = NA,
    alpha = 1,
    linewidth = 1
  ) +
  geom_point(
    data = tibble(),
    aes(x = 0, y = 0),
    color = "black",
    size = 5,
  ) +
  scale_y_reverse() +
  scale_x_reverse()
```

An approach to putting speakers' vowel spaces on a common scale is to divide by the standard deviation.

```{r}
speaker_data_scaled <- speaker_data |>
  mutate(
    .by = speaker,
    across(
      F1:F3,
      ~ .x / sd(.x, na.rm = T)
    )
  )
```

The result is two vowel spaces with center points in the same location, and on a common scale.

```{r}
#| eval: !expr 'rlang::is_installed(c("ggplot2", "ggdensity"))'
#| fig-align: center
#| out-width: 80%
#| code-fold: true
#| code-summary: Plotting code
speaker_data_scaled |>
  mutate(
    .by = speaker,
    across(
      F1:F3,
      ~ (.x - mean(.x, na.rm = T)) / sd(.x, na.rm = T)
    )
  ) |>
  ggplot(
    aes(F2, F1, color = speaker)
  ) +
  stat_hdr(
    probs = c(0.95, 0.8, 0.5),
    fill = NA,
    alpha = 1,
    linewidth = 1
  ) +
  geom_point(
    data = tibble(),
    aes(x = 0, y = 0),
    color = "black",
    size = 5,
  ) +
  scale_y_reverse() +
  scale_x_reverse()
```

# Generalizing Normalization Procedures

All speaker normalization techniques involves some form of Location shift ($L$), Scaling ($S$), or both.
If $F$ is an unnormalized formant value, and $\hat{F}$ is the normalized value, then:

$$
\hat{F} = \frac{F-L}{S}
$$

The primary differences between normalization methods are:

-   The use of across-the-board data transformations.

-   The calculation function for $L$ and $S$.

-   The scope of calculating $L$ and $S$.

The example above is also known as Lobanov Normalization (see `norm_lobanov()`).
The `norm_generic()` function allows for flexible definitions of normalization procedures with this in mind.

## Examples

#### $\Delta$F

The $\Delta$F method was described in @johnsonDFMethodVocal2020.

-   It uses no across-the-board data transformation.

-   It uses a scaling factor $S$ called $\Delta$F, which is the mean across all formants of the formant frequency in Hz, divided by the formant number minus 0.5.

-   $S$ is calculated across all formants

$$
S = \Delta F = \text{mean}\left(\frac{F_{ij}}{i-0.5}\right)
$$

This can be expressed with `norm_generic()` like so

```{r}
speaker_data |>
  norm_generic(
    F1:F3,
    .by = speaker,
    .by_formant = FALSE,
    .by_token = FALSE,
    .L = 0,
    .S = mean(
      .formant / (.formant_num - 0.5),
      na.rm = T
    ),
    .names = "{.formant}_df"
  ) ->
speaker_df_norm
```

```{r}
ggplot(
  speaker_df_norm,
  aes(F2_df, F1_df, color = speaker)
) +
  stat_hdr(
    probs = c(0.95, 0.8, 0.5),
    fill = NA,
    alpha = 1,
    linewidth = 1
  ) +
  scale_y_reverse() +
  scale_x_reverse()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dct.R
\name{reframe_with_dct}
\alias{reframe_with_dct}
\title{Reframe with DCT}
\usage{
reframe_with_dct(
  .data,
  ...,
  .token_id_col = NULL,
  .by = NULL,
  .time_col = NULL,
  .order = 5
)
}
\arguments{
\item{.data}{A data frame}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant.}

\item{.token_id_col}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} The token ID column.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A grouping column.}

\item{.time_col}{A time column.}

\item{.order}{The number of DCT parameters to return. If \code{NA}, all DCT
parameters will be returned.}
}
\value{
A data frame with with the targeted DCT coefficients, along with two
additional columns

\describe{
\item{.param}{The nth DCT coefficient number}
\item{.n}{The number of original data values}
}
}
\description{
Reframe data columns using the Discrete Cosine Transform
}
\details{
This function will tidily apply the Discrete Cosine Transform with forward
normalization (see \link{dct} for more info) to the targeted columns.
\subsection{Identifying tokens}{

The DCT only works on a by-token basis, so there must be a column that
uniquely identifies (or, in combination with a \code{.by} grouping, uniquely
identifies) each individual token. This column should be passed to
\code{.token_id_col}.
}

\subsection{Order}{

The number of DCT coefficients to return is defined by \code{.order}. The default
value is 5. Larger numbers will lead to less smoothing when the Inverse
DCT is applied (see \link{idct}). Smaller numbers will lead to more smoothing.

If \code{NA} is passed to \code{.order}, all DCT parameters will be returned, which
when the Inverse DCT is supplied, will completely reconstruct the original
data.
}

\subsection{Sorting by Time}{

An optional \code{.time_col} can also be defined to ensure that the data is
correctly arranged by time.
}
}
\examples{
library(tidynorm)
library(dplyr)

speaker_small <- filter(
  speaker_tracks,
  id == 0
)

speaker_dct <- reframe_with_dct(
  speaker_small,
  F1:F3,
  .by = speaker,
  .token_id_col = id,
  .time_col = t
)

head(
  speaker_dct
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms.R
\name{hz_to_bark}
\alias{hz_to_bark}
\title{Hz to Bark}
\usage{
hz_to_bark(hz)
}
\arguments{
\item{hz}{Frequency in Hz}
}
\value{
A vector of bark scaled values
}
\description{
Converts Hz to Bark
}
\details{
\deqn{
\hat{b} = \frac{26.81 hz}{1960 + hz} - 0.53
}
\deqn{
b = \begin{cases}
\hat{b} + 0.15(2-\hat{b}) & \text{if}~\hat{b} < 2\\
\hat{b} + 0.22(\hat{b} - 20.1) & \text{if}~\hat{b} > 20.1\\
\hat{b} & \text{otherwise}
\end{cases}
}
}
\examples{
hz <- seq(150, 2000, length = 100)
bark <- hz_to_bark(hz)
plot(hz, bark)

}
\references{
Traunmüller, H. (1990). Analytical expressions for the tonotopic
sensory scale. The Journal of the Acoustical Society of America,
88(1), 97–100. \doi{10.1121/1.399849}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dct.R
\name{reframe_with_dct_smooth}
\alias{reframe_with_dct_smooth}
\title{Reframe with DCT Smooth}
\usage{
reframe_with_dct_smooth(
  .data,
  ...,
  .token_id_col,
  .by = NULL,
  .time_col = NULL,
  .order = 5,
  .rate = FALSE,
  .accel = FALSE
)
}
\arguments{
\item{.data}{A data frame}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant.}

\item{.token_id_col}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} The token ID column.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A grouping column.}

\item{.time_col}{A time column.}

\item{.order}{The number of DCT parameters to return. If \code{NA}, all DCT
parameters will be returned.}

\item{.rate}{Whether or not to include the rate of change of signal.}

\item{.accel}{Whether or not to include acceleration of signal.}
}
\value{
A data frame where the target columns have been smoothed using the
DCT, as well as the signal rate of change and acceleration,
if requested.
}
\description{
Apply a DCT Smooth to the targeted data
}
\details{
This is roughly equivalent to applying \link{reframe_with_dct} followed by
\link{reframe_with_idct}. As long as the value passed to \code{.order} is less than
the length of the each token's data, this will result in a smoothed version
of the data.
\subsection{Identifying tokens}{

The DCT only works on a by-token basis, so there must be a column that
uniquely identifies (or, in combination with a \code{.by} grouping, uniquely
identifies) each individual token. This column should be passed to
\code{.token_id_col}.
}

\subsection{Order}{

The number of DCT coefficients to return is defined by \code{.order}. The default
value is 5. Larger numbers will lead to less smoothing when the Inverse
DCT is applied (see \link{idct}). Smaller numbers will lead to more smoothing.

If \code{NA} is passed to \code{.order}, all DCT parameters will be returned, which
when the Inverse DCT is supplied, will completely reconstruct the original
data.
}

\subsection{Sorting by Time}{

An optional \code{.time_col} can also be defined to ensure that the data is
correctly arranged by time.

Additionally, if \code{.time_col} is provided, the original time column will
be included in the output
}
}
\examples{
library(tidynorm)
library(dplyr)

ggplot2_inst <- require(ggplot2)

speaker_small <- filter(
  speaker_tracks,
  id == 0
)

speaker_dct_smooth <- speaker_small |>
  reframe_with_dct_smooth(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .order = 5
  )

if (ggplot2_inst) {
  speaker_small |>
    ggplot(
      aes(
        t, F1
      )
    ) +
    geom_point() +
    facet_wrap(
      ~speaker,
      scales = "free_x",
      ncol = 1
    ) +
    labs(
      title = "Original Data"
    )
}

if (ggplot2_inst) {
  speaker_dct_smooth |>
    ggplot(
      aes(
        t, F1
      )
    ) +
    geom_point() +
    facet_wrap(
      ~speaker,
      scales = "free_x",
      ncol = 1
    ) +
    labs(
      title = "Smoothed Data"
    )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{tidynorm_options}
\alias{tidynorm_options}
\title{Set tidynorm options}
\usage{
tidynorm_options(
  .silent = opt("tidynorm.silent"),
  .warnings = opt("tidynorm.warnings")
)
}
\arguments{
\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}

\item{.warnings}{Print warnings from tidynorm functions. (Defaults to \code{TRUE}, overwritable using option 'tidynorm.warnings' or environment variable 'R_TIDYNORM_TIDYNORM_WARNINGS')}
}
\description{
Set tidynorm verbosity
}
\examples{
tidynorm_options(.silent = TRUE, .warnings = FALSE)

speaker_data |>
  norm_generic(F1:F3) ->
  norm1

tidynorm_options(.silent = FALSE, .warnings = TRUE)

speaker_data |>
  norm_generic(F1:F3) ->
  norm2

}
\seealso{
\link{options}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/track_norm.R
\name{norm_track_barkz}
\alias{norm_track_barkz}
\title{Bark Difference Track Normalization}
\usage{
norm_track_barkz(
  .data,
  ...,
  .token_id_col,
  .by = NULL,
  .time_col = NULL,
  .order = 5,
  .return_dct = FALSE,
  .drop_orig = FALSE,
  .names = "{.formant}_bz",
  .silent = opt("tidynorm.silent")
)
}
\arguments{
\item{.data}{A data frame containing vowel formant data}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant
data columns.}

\item{.token_id_col}{\code{\link[rlang:args_data_masking]{<data-masking>}} A column
that identifies token ids.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A selection of
columns to group by. Typically a column of speaker IDs.}

\item{.time_col}{\code{\link[rlang:args_data_masking]{<data-masking>}} A time column.
(optional)}

\item{.order}{The number of DCT parameters to use.}

\item{.return_dct}{Whether or not the normalized DCT coefficients themselves
should be returned.}

\item{.drop_orig}{Should the originally targeted columns be dropped.}

\item{.names}{A \code{\link[glue:glue]{glue::glue()}} expression for naming the normalized
data columns. The \code{"{.formant}"} portion corresponds to the name of the original
formant columns.}

\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}
}
\value{
A data frame of either normalized formant tracks, or normalized DCT
parameters.

A data frame of Bark difference normalized formant tracks.
}
\description{
Bark Difference Track Normalization
}
\details{
This is a within-token normalization technique. First all formants are
converted to Bark (see \link{hz_to_bark}), then, within each token, F3 is
subtracted from F1 and F2.

\deqn{
\hat{F}_{ij} = F_{ij} - L_j
}

\deqn{
L_j = F_{3j}
}
}
\examples{
library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

track_subset <- speaker_tracks |>
  filter(
    .by = c(speaker, id),
    if_all(
      F1:F3,
      .fns = \(x) mean(is.finite(x)) > 0.9
    ),
    row_number() \%\% 2 == 1
  )

track_norm <- track_subset |>
  norm_track_barkz(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .drop_orig = TRUE
  )

if (ggplot2_inst) {
  track_norm |>
    ggplot(
      aes(F2_bz, F1_bz, color = speaker)
    ) +
    stat_density_2d(bins = 4) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}


# returning the DCT coefficients
track_norm_dct <- track_subset |>
  norm_track_barkz(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .drop_orig = TRUE,
    .return_dct = TRUE,
    .names = "{.formant}_bz"
  )

track_norm_means <- track_norm_dct |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      ends_with("_bz"),
      mean
    )
  ) |>
  reframe_with_idct(
    ends_with("_bz"),
    .by = speaker,
    .token_id_col = vowel,
    .param_col = .param
  )


if (ggplot2_inst) {
  track_norm_means |>
    ggplot(
      aes(F2_bz, F1_bz, color = speaker)
    ) +
    geom_path(
      aes(
        group = interaction(speaker, vowel)
      )
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}
}
\references{
Syrdal, A. K., & Gopal, H. S. (1986). A perceptual model of vowel
recognition based on the auditory representation of American English vowels.
The Journal of the Acoustical Society of America, 79(4), 1086–1100.
\doi{10.1121/1.393381}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/norm.R
\name{norm_wattfab}
\alias{norm_wattfab}
\title{Watt & Fabricius Normalize}
\usage{
norm_wattfab(
  .data,
  ...,
  .by = NULL,
  .by_formant = TRUE,
  .drop_orig = FALSE,
  .keep_params = FALSE,
  .names = "{.formant}_wf",
  .silent = opt("tidynorm.silent")
)
}
\arguments{
\item{.data}{A data frame containing vowel formant data}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant
data columns.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A selection of
columns to group by. Typically a column of speaker IDs.}

\item{.by_formant}{Ignored by this procedure}

\item{.drop_orig}{Whether or not to drop the original formant data columns.}

\item{.keep_params}{Whether or not to keep the Location (\verb{*_.L}) and Scale
(\verb{*_.S}) normalization parameters}

\item{.names}{A \code{\link[glue:glue]{glue::glue()}} expression for naming the normalized
data columns. The \code{"{.formant}"} portion corresponds to the name of the original
formant columns.}

\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}
}
\value{
A data fame of Watt & Fabricius normalized formant values.
}
\description{
Watt & Fabricius Normalize
}
\details{
This is a modified version of the Watt & Fabricius Method. The original
method identified point vowels over which F1 and F2 centroids were calculated.
The procedure here just identifies centroids by taking the mean of
all formant values.

\deqn{
\hat{F}_{ij} = \frac{F_{ij}}{S_i}
}

\deqn{
 S_i = \frac{1}{N}\sum_{j=1}^N F_{ij}
}

Where
\itemize{
\item \eqn{\hat{F}} is the normalized formant
\item \eqn{i} is the formant number
\item \eqn{j} is the token number
}
}
\examples{
library(tidynorm)
ggplot2_inst <- require(ggplot2)

speaker_data_wattfab <- speaker_data |>
  norm_wattfab(
    F1:F3,
    .by = speaker,
    .names = "{.formant}_wf"
  )

if (ggplot2_inst) {
  ggplot(
    speaker_data_wattfab,
    aes(
      F2_wf,
      F1_wf,
      color = speaker
    )
  ) +
    stat_density_2d(
      bins = 4
    ) +
    scale_color_brewer(
      palette = "Dark2"
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    coord_fixed()
}
}
\references{
Watt, D., & Fabricius, A. (2002). Evaluation of a technique for improving the
mapping of multiple speakers’ vowel spaces in the F1 ~ F2 plane.
Leeds Working Papers in Linguistics and Phonetics, 9, 159–173.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dct_norm.R
\name{norm_dct_generic}
\alias{norm_dct_generic}
\title{Generic Formant DCT Normalization Procedure}
\usage{
norm_dct_generic(
  .data,
  ...,
  .token_id_col,
  .by = NULL,
  .param_col = NULL,
  .L = 0,
  .S = 1/sqrt(2),
  .by_formant = FALSE,
  .by_token = FALSE,
  .names = "{.formant}_n",
  .silent = opt("tidynorm.silent"),
  .drop_orig = FALSE,
  .call = caller_env()
)
}
\arguments{
\item{.data}{A data frame of formant DCT coefficients}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant
data columns.}

\item{.token_id_col}{\code{\link[rlang:args_data_masking]{<data-masking>}} A column
that identifies token ids.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A selection of
columns to group by. Typically a column of speaker IDs.}

\item{.param_col}{A column identifying the DCT parameter number.}

\item{.L}{An expression defining the location parameter.
See Details for more information.}

\item{.S}{An expression defining the scale parameter.
See Details for more information.}

\item{.by_formant}{Whether or not the normalization method is formant
intrinsic.}

\item{.by_token}{Whether or not the normalization method is token
intrinsic}

\item{.names}{A \code{\link[glue:glue]{glue::glue()}} expression for naming the normalized
data columns. The \code{"{.formant}"} portion corresponds to the name of the original
formant columns.}

\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}

\item{.drop_orig}{Should the originally targeted columns be dropped.}

\item{.call}{Used for internal purposes.}
}
\value{
A data frame of normalized DCT coefficients.
}
\description{
Generic Formant DCT Normalization Procedure
}
\details{
The following \verb{norm_dct_*} procedures were built on top of
\code{norm_dct_generic()}.
\itemize{
\item \link{norm_dct_lobanov}
\item \link{norm_dct_nearey}
\item \link{norm_dct_deltaF}
\item \link{norm_dct_wattfab}
\item \link{norm_dct_barkz}
}
\subsection{Normalizing DCT Coefficients}{

This will normalize vowel formant data that has already had the
Discrete Cosine Transform applied (see \link{dct}) with the following
procedure:
\enumerate{
\item Location \code{.L} and Scale \code{.S} expressions will be used to summarize
the zero\if{html}{\out{<sup>}}th\if{html}{\out{</sup>}} DCT coefficients.
\item These location and scale will be used to normalize the DCT coefficients.
}
\subsection{Location and Scale expressions}{

\link{norm_dct_generic} normalizes DCT coefficients directly.
If \eqn{F_k} is the k\if{html}{\out{<sup>}}th\if{html}{\out{</sup>}} DCT coefficient
the normalization procedure is

\deqn{
\hat{F}_k = \frac{F_k - L'}{\sqrt{2}S}
}
\deqn{
L' = \begin{cases}
   L & \text{for }k=0\\
   0 & \text{for }k>0
 \end{cases}
}

Rather than requiring users to remember to multiply expressions for \eqn{S}
by \eqn{\sqrt{2}}, this is done by \link{norm_dct_generic} itself, to allow greater
parallelism with how \link{norm_generic} works.

\strong{Note}: If you want to scale values by a constant in the normalization,
you'll need to divide the constant by \code{sqrt(2)}.

The expressions for calculating \eqn{L} and \eqn{S} can be
passed to \code{.L} and \code{.S}, respectively. Available values for
these expressions are

\describe{
\item{\code{.formant}}{The original formant value}
\item{\code{.formant_num}}{The number of the formant. (e.g. 1 for F1, 2 for F2 etc)}
}

Along with any data columns from your original data.
}

\subsection{Identifying tokens}{

DCT normalization requires identifying individual tokens, so there must be a column that
uniquely identifies (or, in combination with a \code{.by} grouping, uniquely
identifies) each individual token. This column should be passed to
\code{.token_id_col}.
}

}
}
\examples{
library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

track_subset <- speaker_tracks |>
  filter(
    .by = c(speaker, id),
    if_all(
      F1:F3,
      .fns = \(x) mean(is.finite(x)) > 0.9
    ),
    row_number() \%\% 2 == 1
  )

track_dcts <- track_subset |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .order = 3
  )

track_norm <- track_dcts |>
  norm_dct_generic(
    F1:F3,
    .token_id_col = id,
    .by = speaker,
    .by_formant = TRUE,
    .L = median(.formant, na.rm = TRUE),
    .S = mad(.formant, na.rm = TRUE),
    .param_col = .param,
    .drop_orig = TRUE,
    .names = "{.formant}_mad"
  )

head(track_norm)

full_tracks <- track_norm |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      F1_mad:F3_mad,
      mean
    )
  ) |>
  reframe_with_idct(
    F1_mad:F3_mad,
    .by = c(speaker, vowel),
    .param_col = .param
  )

head(full_tracks)

if (ggplot2_inst) {
  ggplot(
    full_tracks,
    aes(F2_mad, F1_mad, color = speaker)
  ) +
    geom_path(
      aes(group = interaction(speaker, vowel))
    ) +
    scale_y_reverse() +
    scale_x_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dct.R
\name{idct_rate}
\alias{idct_rate}
\title{Inverse Discrete Cosine Transform Rate}
\usage{
idct_rate(y, n = length(y))
}
\arguments{
\item{y}{DCT coefficients}

\item{n}{The desired length of the idct}
}
\value{
A vector with the first derivative
of the inverse DCT
}
\description{
The first derivative of the Inverse DCT
}
\details{
Returns the first derivative (rate of change) of
the Inverse DCT (see \link{dct} for more details).

\deqn{
\frac{\delta x_j}{\delta j} = -2\frac{\pi k}{J}\sum_{k=1}^{N-1} y_k \sin\left(\frac{\pi k(2j+1)}{2J}\right)
}
}
\examples{
x <- seq(0, 1, length = 10)
y <- 5 + x + (2 * (x^2)) + (-2 * (x^4))

dct_coefs <- dct(y)
y_rate <- idct_rate(dct_coefs)

plot(y)
plot(y_rate)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dct.R
\name{dct}
\alias{dct}
\alias{dct.numeric}
\alias{dct.matrix}
\title{Discrete Cosine Transform}
\usage{
dct(x)

\method{dct}{numeric}(x)

\method{dct}{matrix}(x)
}
\arguments{
\item{x}{A vector or matrix to which the discrete cosine transform is applied}
}
\value{
Returned value depends on \code{x}.

When passed a numeric vector, returns a numeric
vector the same size as \code{x} with the DCT Coefficients.

When passed a matrix, returns a matrix
the same size as \code{x} with the DCT Coefficients.
}
\description{
Discrete Cosine Transform
}
\details{
The DCT definitions here are based on the python \code{scipy.fft.dct} definitions.
Specifically this use:

\if{html}{\out{<div class="sourceCode python">}}\preformatted{# python code
scipy.fft.dct(x, norm = "forward", orthogonalize = True)
}\if{html}{\out{</div>}}

\deqn{
y_k = \frac{1}{zN} \sum_{j=0}^{N-1}x_j\cos\left(\frac{\pi k(2j+1)}{2N}\right)
}

\deqn{
z = \begin{cases}
   \sqrt{2}& \text{for }k=0\\
   1 & \text{for }k>0
 \end{cases}
}

For the Inverse Discrete Cosine Transform, see \link{idct}.
}
\examples{
x <- seq(0, 1, length = 10)
y <- 5 + x + (2 * (x^2)) + (-2 * (x^4))

dct_coefs <- dct(y)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/track_norm.R
\name{norm_track_nearey}
\alias{norm_track_nearey}
\title{Nearey Track Normalization}
\usage{
norm_track_nearey(
  .data,
  ...,
  .token_id_col,
  .by = NULL,
  .by_formant = FALSE,
  .time_col = NULL,
  .order = 5,
  .return_dct = FALSE,
  .drop_orig = FALSE,
  .names = "{.formant}_lm",
  .silent = opt("tidynorm.silent")
)
}
\arguments{
\item{.data}{A data frame containing vowel formant data}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant
data columns.}

\item{.token_id_col}{\code{\link[rlang:args_data_masking]{<data-masking>}} A column
that identifies token ids.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A selection of
columns to group by. Typically a column of speaker IDs.}

\item{.by_formant}{Whether or not the normalization method is formant
intrinsic.}

\item{.time_col}{\code{\link[rlang:args_data_masking]{<data-masking>}} A time column.
(optional)}

\item{.order}{The number of DCT parameters to use.}

\item{.return_dct}{Whether or not the normalized DCT coefficients themselves
should be returned.}

\item{.drop_orig}{Should the originally targeted columns be dropped.}

\item{.names}{A \code{\link[glue:glue]{glue::glue()}} expression for naming the normalized
data columns. The \code{"{.formant}"} portion corresponds to the name of the original
formant columns.}

\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}
}
\value{
A data frame of Nearey normalized formant tracks.
}
\description{
Nearey Track Normalization
}
\details{
When formant extrinsic:
\deqn{
 \hat{F}_{ij} = \log(F_{ij}) - L
}
\deqn{
 L = \frac{1}{MN}\sum_{i=1}^M\sum_{j=1}^N \log(F_{ij})
}

When formant intrinsic:
\deqn{
 \hat{F}_{ij} = \log(F_{ij}) - L_{i}
}

\deqn{
  L_i = \frac{1}{N}\sum_{j=1}^{N}\log(F_{ij})
}

Where
\itemize{
\item \eqn{\hat{F}} is the normalized formant
\item \eqn{i} is the formant number
\item \eqn{j} is the token number
}
}
\examples{
library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

track_subset <- speaker_tracks |>
  filter(
    .by = c(speaker, id),
    if_all(
      F1:F3,
      .fns = \(x) mean(is.finite(x)) > 0.9
    ),
    row_number() \%\% 2 == 1
  )

track_norm <- track_subset |>
  norm_track_nearey(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .by_formant = TRUE,
    .drop_orig = TRUE
  )

if (ggplot2_inst) {
  track_norm |>
    ggplot(
      aes(F2_lm, F1_lm, color = speaker)
    ) +
    stat_density_2d(bins = 4) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}


# returning the DCT coefficients
track_norm_dct <- track_subset |>
  norm_track_nearey(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .by_formant = FALSE,
    .drop_orig = TRUE,
    .return_dct = TRUE,
    .names = "{.formant}_lm"
  )

track_norm_means <- track_norm_dct |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      ends_with("_lm"),
      mean
    )
  ) |>
  reframe_with_idct(
    ends_with("_lm"),
    .by = speaker,
    .token_id_col = vowel,
    .param_col = .param
  )


if (ggplot2_inst) {
  track_norm_means |>
    ggplot(
      aes(F2_lm, F1_lm, color = speaker)
    ) +
    geom_path(
      aes(
        group = interaction(speaker, vowel)
      )
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}
}
\references{
Nearey, T. M. (1978). Phonetic Feature Systems for Vowels [Ph.D.].
University of Alberta.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidynorm-package.R
\docType{package}
\name{tidynorm-package}
\alias{tidynorm}
\alias{tidynorm-package}
\title{tidynorm: Tools for Tidy Vowel Normalization}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

An implementation of tidy speaker vowel normalization. This includes generic functions for defining new normalization methods for points, formant tracks, and Discrete Cosine Transform coefficients, as well as convenience functions implementing established normalization methods. References for the implemented methods are: Johnson, Keith (2020) \doi{10.5334/labphon.196} Lobanov, Boris (1971) \doi{10.1121/1.1912396} Nearey, Terrance M. (1978) \url{https://sites.ualberta.ca/~tnearey/Nearey1978_compressed.pdf} Syrdal, Ann K., and Gopal, H. S. (1986) \doi{10.1121/1.393381} Watt, Dominic, and Fabricius, Anne (2002) \url{https://www.latl.leeds.ac.uk/article/evaluation-of-a-technique-for-improving-the-mapping-of-multiple-speakers-vowel-spaces-in-the-f1-f2-plane/}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://jofrhwld.github.io/tidynorm/}
  \item \url{https://github.com/JoFrhwld/tidynorm}
  \item Report bugs at \url{https://github.com/JoFrhwld/tidynorm/issues}
}

}
\author{
\strong{Maintainer}: Josef Fruehwald \email{JoFrhwld@gmail.com} [copyright holder]

}
\keyword{##}
\keyword{internal}
\keyword{namespace:}
\keyword{start}
\keyword{usethis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dct.R
\name{dct_basis}
\alias{dct_basis}
\title{DCT Basis}
\usage{
dct_basis(n, k)
}
\arguments{
\item{n}{The length of the basis.}

\item{k}{The number of basis functions.}
}
\value{
A \eqn{n\times k} matrix
}
\description{
The Discrete Cosine Transform basis functions
}
\details{
This function will generate the DCT basis functions.
}
\examples{
basis <- dct_basis(100, 5)
matplot(basis, type = "l", lty = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_norm}
\alias{check_norm}
\title{Check Normalization Procedures}
\usage{
check_norm(.data)
}
\arguments{
\item{.data}{A data frame produced by a tidynorm function.}
}
\value{
This only prints an info message.
}
\description{
\code{check_norm()} will generate a message with information
about which normalization procedures have been applied to the
data.
}
\examples{
speaker_norm <- speaker_data |>
  norm_nearey(
    F1:F3,
    .by = speaker,
    .silent = TRUE
  )

check_norm(speaker_norm)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dct.R
\name{reframe_with_idct}
\alias{reframe_with_idct}
\title{Reframe with IDCT}
\usage{
reframe_with_idct(
  .data,
  ...,
  .token_id_col = NULL,
  .by = NULL,
  .param_col = NULL,
  .n = 20,
  .rate = FALSE,
  .accel = FALSE
)
}
\arguments{
\item{.data}{A data frame}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant.}

\item{.token_id_col}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} The token ID column.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A grouping column.}

\item{.param_col}{A column identifying the DCT parameter number}

\item{.n}{The size of the outcome of the IDCT}

\item{.rate}{Whether or not to include the rate of change of signal.}

\item{.accel}{Whether or not to include acceleration of signal.}
}
\value{
A data frame with the IDCT of the targeted columns along with an
additional \code{.time} column.

\describe{
\item{.time}{A column from 1 to \code{.n} by token}
}
}
\description{
Reframe data columns using the Inverse Discrete Cosine Transform
}
\details{
This will apply the Inverse Discrete Cosine Transform to the targeted
columns. See \link{idct}.
\subsection{Identifying tokens}{

The IDCT only works on a by-token basis, so there must be a column that
uniquely identifies (or, in combination with a \code{.by} grouping, uniquely
identifies) each individual token. This column should be passed to
\code{.token_id_col}.
}

\subsection{Size of the output}{

The output of the IDCT can be arbitrarily long as defined by the \code{.n}
argument. \code{.n} can either be an integer, or an unqoted data column.
}

\subsection{The Parameter Column}{

The order of the DCT parameters is crucially important. The optional
\code{.param_col} will ensure the data is properly arranged.
}
}
\examples{
library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

speaker_small <- filter(
  speaker_tracks,
  id == 0
)

speaker_dct <- speaker_small |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t,
    .order = 5
  )

speaker_idct <- speaker_dct |>
  reframe_with_idct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param,
    .n = 20
  )

if (ggplot2_inst) {
  speaker_small |>
    mutate(
      .by = c(speaker, id),
      time_index = row_number()
    ) |>
    ggplot(
      aes(
        time_index, F1
      )
    ) +
    geom_point() +
    labs(
      title = "Original Data"
    )
}

if (ggplot2_inst) {
  speaker_idct |>
    ggplot(
      aes(
        .time, F1
      )
    ) +
    geom_point() +
    labs(
      title = "DCT Smooth Data"
    )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/norm.R
\name{norm_generic}
\alias{norm_generic}
\title{Generic Normalization Procedure}
\usage{
norm_generic(
  .data,
  ...,
  .by = NULL,
  .by_formant = FALSE,
  .by_token = FALSE,
  .L = 0,
  .S = 1,
  .pre_trans = identity,
  .post_trans = identity,
  .drop_orig = FALSE,
  .keep_params = FALSE,
  .names = "{.formant}_n",
  .silent = opt("tidynorm.silent"),
  .call = caller_env()
)
}
\arguments{
\item{.data}{A data frame containing vowel formant data}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant
data columns.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A selection of
columns to group by. Typically a column of speaker IDs.}

\item{.by_formant}{Whether or not the normalization method is formant
intrinsic.}

\item{.by_token}{Whether or not the normalization method is vowel intrinsic}

\item{.L}{An expression defining the location parameter.
See Details for more information.}

\item{.S}{An expression defining the scale parameter.
See Details for more information.}

\item{.pre_trans}{A function to apply to formant values before normalization.}

\item{.post_trans}{A function to apply to formant values after normalization.}

\item{.drop_orig}{Whether or not to drop the original formant data columns.}

\item{.keep_params}{Whether or not to keep the Location (\verb{*_.L}) and Scale
(\verb{*_.S}) normalization parameters}

\item{.names}{A \code{\link[glue:glue]{glue::glue()}} expression for naming the normalized
data columns. The \code{"{.formant}"} portion corresponds to the name of the original
formant columns.}

\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}

\item{.call}{Used for internal purposes.}
}
\value{
A data frame of normalized formant values
}
\description{
This is a generic normalization procedure with which
you can create your own normalization method.
}
\details{
The following \verb{norm_*} procedures are built on top of \code{norm_generic()}.
\itemize{
\item \link{norm_lobanov}
\item \link{norm_nearey}
\item \link{norm_deltaF}
\item \link{norm_wattfab}
\item \link{norm_barkz}
}
\subsection{Location and Scale expressions}{

All normalization procedures built on \link{norm_generic} produce normalized
formant values (\eqn{\hat{F}}) by subtracting a location parameter
(\eqn{L}) and dividing by a scale parameter (\eqn{S}).

\deqn{
\hat{F} = \frac{F-L}{S}
}

The expressions for calculating \eqn{L} and \eqn{S} can be
passed to \code{.L} and \code{.S}, respectively. Available values for
these expressions are

\describe{
\item{\code{.formant}}{The original formant value}
\item{\code{.formant_num}}{The number of the formant. (e.g. 1 for F1, 2 for F2 etc)}
}

Along with any data columns from your original data.
}

\subsection{Pre and Post normalization transforms}{

To apply any transformations before or after normalization,
you can pass a function to \code{.pre_trans} and \code{.post_trans}.
}

\subsection{Formant In/Extrinsic Normalization}{

If \code{.by_formant} is \code{TRUE}, normalization will be formant intrinsic.
If \code{.by_formant} is \code{FALSE}, normalization will be formant extrinsic.
}

\subsection{Token In/Extrinsic Normalization}{

If \code{.by_token} is \code{TRUE}, normalization will be token intrinsic.
If \code{.by_token} is \code{FALSE}, normalization will be token extrinsic.
}
}
\examples{
library(tidynorm)
library(dplyr)

speaker_data |>
  norm_generic(
    F1:F3,
    .by = speaker,
    .by_formant = TRUE,
    .L = median(.formant, na.rm = TRUE),
    .S = mad(.formant, na.rm = TRUE),
    .drop_orig = TRUE,
    .names = "{.formant}_mad"
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/norm.R
\name{norm_barkz}
\alias{norm_barkz}
\title{Bark Difference Normalize}
\usage{
norm_barkz(
  .data,
  ...,
  .by = NULL,
  .drop_orig = FALSE,
  .keep_params = FALSE,
  .names = "{.formant}_bz",
  .silent = opt("tidynorm.silent")
)
}
\arguments{
\item{.data}{A data frame containing vowel formant data}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant
data columns.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A selection of
columns to group by. Typically a column of speaker IDs.}

\item{.drop_orig}{Whether or not to drop the original formant data columns.}

\item{.keep_params}{Whether or not to keep the Location (\verb{*_.L}) and Scale
(\verb{*_.S}) normalization parameters}

\item{.names}{A \code{\link[glue:glue]{glue::glue()}} expression for naming the normalized
data columns. The \code{"{.formant}"} portion corresponds to the name of the original
formant columns.}

\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}
}
\value{
A data frame of Bark Difference normalized formant values
}
\description{
Bark Difference Normalize
}
\details{
This is a within-token normalization technique. First all formants are
converted to Bark (see \link{hz_to_bark}), then, within each token, F3 is
subtracted from F1 and F2.

\deqn{
\hat{F}_{ij} = F_{ij} - L_j
}

\deqn{
L_j = F_{3j}
}
}
\examples{
library(tidynorm)
ggplot2_inst <- require(ggplot2)

speaker_data_barkz <- speaker_data |>
  norm_barkz(
    F1:F3,
    .by = speaker,
    .names = "{.formant}_bz"
  )

if (ggplot2_inst) {
  ggplot(
    speaker_data_barkz,
    aes(
      F2_bz,
      F1_bz,
      color = speaker
    )
  ) +
    stat_density_2d(
      bins = 4
    ) +
    scale_color_brewer(
      palette = "Dark2"
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    coord_fixed()
}
}
\references{
Syrdal, A. K., & Gopal, H. S. (1986). A perceptual model of vowel
recognition based on the auditory representation of American English vowels.
The Journal of the Acoustical Society of America, 79(4), 1086–1100.
\doi{10.1121/1.393381}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/norm.R
\name{norm_deltaF}
\alias{norm_deltaF}
\title{Delta F Normalize}
\usage{
norm_deltaF(
  .data,
  ...,
  .by = NULL,
  .by_formant = FALSE,
  .drop_orig = FALSE,
  .keep_params = FALSE,
  .names = "{.formant}_df",
  .silent = opt("tidynorm.silent")
)
}
\arguments{
\item{.data}{A data frame containing vowel formant data}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant
data columns.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A selection of
columns to group by. Typically a column of speaker IDs.}

\item{.by_formant}{Ignored by this procedure}

\item{.drop_orig}{Whether or not to drop the original formant data columns.}

\item{.keep_params}{Whether or not to keep the Location (\verb{*_.L}) and Scale
(\verb{*_.S}) normalization parameters}

\item{.names}{A \code{\link[glue:glue]{glue::glue()}} expression for naming the normalized
data columns. The \code{"{.formant}"} portion corresponds to the name of the original
formant columns.}

\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}
}
\value{
A data frame of Delta F normalized formant values.
}
\description{
Delta F Normalize
}
\details{
\deqn{
 \hat{F}_{ij} = \frac{F_{ij}}{S}
}
\deqn{
 S = \frac{1}{MN}\sum_{i=1}^M\sum_{j=1}^N \frac{F_{ij}}{i-0.5}
}

Where
\itemize{
\item \eqn{\hat{F}} is the normalized formant
\item \eqn{i} is the formant number
\item \eqn{j} is the token number
}
}
\examples{
library(tidynorm)
ggplot2_inst <- require(ggplot2)

speaker_data_deltaF <- speaker_data |>
  norm_deltaF(
    F1:F3,
    .by = speaker,
    .names = "{.formant}_df"
  )

if (ggplot2_inst) {
  ggplot(
    speaker_data_deltaF,
    aes(
      F2_df,
      F1_df,
      color = speaker
    )
  ) +
    stat_density_2d(
      bins = 4
    ) +
    scale_color_brewer(
      palette = "Dark2"
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    coord_fixed()
}
}
\references{
Johnson, K. (2020). The \eqn{\Delta}F method of vocal tract length normalization for vowels.
Laboratory Phonology: Journal of the Association for Laboratory Phonology, 11(1),
Article 1. \doi{10.5334/labphon.196}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/norm.R
\name{norm_nearey}
\alias{norm_nearey}
\title{Nearey Normalize}
\usage{
norm_nearey(
  .data,
  ...,
  .by = NULL,
  .by_formant = FALSE,
  .drop_orig = FALSE,
  .keep_params = FALSE,
  .names = "{.formant}_lm",
  .silent = opt("tidynorm.silent")
)
}
\arguments{
\item{.data}{A data frame containing vowel formant data}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant
data columns.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A selection of
columns to group by. Typically a column of speaker IDs.}

\item{.by_formant}{Whether or not the normalization method is formant
intrinsic.}

\item{.drop_orig}{Whether or not to drop the original formant data columns.}

\item{.keep_params}{Whether or not to keep the Location (\verb{*_.L}) and Scale
(\verb{*_.S}) normalization parameters}

\item{.names}{A \code{\link[glue:glue]{glue::glue()}} expression for naming the normalized
data columns. The \code{"{.formant}"} portion corresponds to the name of the original
formant columns.}

\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}
}
\value{
A data frame of Nearey normalized formant values.
}
\description{
Nearey Normalize
}
\details{
When formant extrinsic:
\deqn{
 \hat{F}_{ij} = \log(F_{ij}) - L
}
\deqn{
 L = \frac{1}{MN}\sum_{i=1}^M\sum_{j=1}^N \log(F_{ij})
}

When formant intrinsic:
\deqn{
 \hat{F}_{ij} = \log(F_{ij}) - L_{i}
}

\deqn{
  L_i = \frac{1}{N}\sum_{j=1}^{N}\log(F_{ij})
}

Where
\itemize{
\item \eqn{\hat{F}} is the normalized formant
\item \eqn{i} is the formant number
\item \eqn{j} is the token number
}
}
\examples{
library(tidynorm)
ggplot2_inst <- require(ggplot2)

speaker_data_nearey <- speaker_data |>
  norm_nearey(
    F1:F3,
    .by = speaker,
    .by_formant = FALSE,
    .names = "{.formant}_nearey"
  )

if (ggplot2_inst) {
  ggplot(
    speaker_data_nearey,
    aes(
      F2_nearey,
      F1_nearey,
      color = speaker
    )
  ) +
    stat_density_2d(
      bins = 4
    ) +
    scale_color_brewer(
      palette = "Dark2"
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    coord_fixed() +
    labs(
      title = "Formant extrinsic"
    )
}

speaker_data_nearey2 <- speaker_data |>
  norm_nearey(
    F1:F3,
    .by = speaker,
    .by_formant = TRUE,
    .names = "{.formant}_nearey"
  )

if (ggplot2_inst) {
  ggplot(
    speaker_data_nearey2,
    aes(
      F2_nearey,
      F1_nearey,
      color = speaker
    )
  ) +
    stat_density_2d(
      bins = 4
    ) +
    scale_color_brewer(
      palette = "Dark2"
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    coord_fixed() +
    labs(
      title = "Formant intrinsic"
    )
}
}
\references{
Nearey, T. M. (1978). Phonetic Feature Systems for Vowels [Ph.D.].
University of Alberta.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dct_norm.R
\name{norm_dct_wattfab}
\alias{norm_dct_wattfab}
\title{Watt and Fabricius DCT normalization}
\usage{
norm_dct_wattfab(
  .data,
  ...,
  .token_id_col,
  .by = NULL,
  .param_col = NULL,
  .drop_orig = FALSE,
  .names = "{.formant}_wf",
  .silent = opt("tidynorm.silent")
)
}
\arguments{
\item{.data}{A data frame containing vowel formant data}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant
data columns.}

\item{.token_id_col}{\code{\link[rlang:args_data_masking]{<data-masking>}} A column
that identifies token ids.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A selection of
columns to group by. Typically a column of speaker IDs.}

\item{.param_col}{A column identifying the DCT parameter number.}

\item{.drop_orig}{Should the originally targeted columns be dropped.}

\item{.names}{A \code{\link[glue:glue]{glue::glue()}} expression for naming the normalized
data columns. The \code{"{.formant}"} portion corresponds to the name of the original
formant columns.}

\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}
}
\value{
A data frame of Watt & Fabricius normalized DCT coefficients.
}
\description{
Watt and Fabricius DCT normalization
}
\details{
This is a modified version of the Watt & Fabricius Method. The original
method identified point vowels over which F1 and F2 centroids were calculated.
The procedure here just identifies centroids by taking the mean of
all formant values.

\deqn{
\hat{F}_{ij} = \frac{F_{ij}}{S_i}
}

\deqn{
 S_i = \frac{1}{N}\sum_{j=1}^N F_{ij}
}

Where
\itemize{
\item \eqn{\hat{F}} is the normalized formant
\item \eqn{i} is the formant number
\item \eqn{j} is the token number
}
}
\examples{
library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

speaker_dct <- speaker_tracks |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )

# Normalize DCT coefficients
speaker_dct_norm <- speaker_dct |>
  norm_dct_wattfab(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param
  )

# Apply average and apply inverse dct
# to plot tracks
track_norm_means <- speaker_dct_norm |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      ends_with("_wf"),
      mean
    )
  ) |>
  reframe_with_idct(
    ends_with("_wf"),
    .by = speaker,
    .token_id_col = vowel,
    .param_col = .param
  )


if (ggplot2_inst) {
  track_norm_means |>
    ggplot(
      aes(F2_wf, F1_wf, color = speaker)
    ) +
    geom_path(
      aes(
        group = interaction(speaker, vowel)
      )
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}
}
\references{
Watt, D., & Fabricius, A. (2002). Evaluation of a technique for improving the
mapping of multiple speakers’ vowel spaces in the F1 ~ F2 plane.
Leeds Working Papers in Linguistics and Phonetics, 9, 159–173.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{speaker_data}
\alias{speaker_data}
\title{Speaker Data}
\format{
\subsection{\code{speaker_data}}{

A data frame with 10,697 rows and 8 columns
\describe{
\item{speaker}{Speaker ID column}
\item{vowel}{CMU Dictionary vowel class}
\item{plt_vclass}{Modified Labov-Trager vowel class}
\item{ipa_vclas}{IPA-like vowel class}
\item{word}{Word that the vowel appeared in}
\item{F1, F2, F3}{The first, second and third formants, in Hz}
}
}
}
\usage{
speaker_data
}
\description{
Speaker Data
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms.R
\name{bark_to_hz}
\alias{bark_to_hz}
\title{Bark to Hz}
\usage{
bark_to_hz(bark)
}
\arguments{
\item{bark}{Frequency in Bark}
}
\value{
A vector of Hz scaled values
}
\description{
Converts bark to Hz
}
\details{
\deqn{
\hat{b} = \begin{cases}
\frac{b - 0.3}{0.85} & \text{if} ~ b < 2\\
\frac{b + 4.422}{1.22} & \text{if} ~ b > 20.1\\
b & \text{otherwise}
\end{cases}
}

\deqn{
hz = 1960\frac{\hat{b} + 0.53}{26.28 - \hat{b}}
}
}
\examples{
bark <- seq(1.5, 13, length = 100)
hz <- bark_to_hz(bark)
plot(bark, hz)
}
\references{
Traunmüller, H. (1990). Analytical expressions for the tonotopic
sensory scale. The Journal of the Acoustical Society of America,
88(1), 97–100. \doi{10.1121/1.399849}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dct_norm.R
\name{norm_dct_nearey}
\alias{norm_dct_nearey}
\title{Nearey DCT Normalization}
\usage{
norm_dct_nearey(
  .data,
  ...,
  .token_id_col,
  .by = NULL,
  .by_formant = FALSE,
  .param_col = NULL,
  .drop_orig = FALSE,
  .names = "{.formant}_lm",
  .silent = opt("tidynorm.silent")
)
}
\arguments{
\item{.data}{A data frame containing vowel formant data}

\item{...}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} One or more unquoted
expressions separated by commas. These should target the vowel formant
data columns.}

\item{.token_id_col}{\code{\link[rlang:args_data_masking]{<data-masking>}} A column
that identifies token ids.}

\item{.by}{\code{\link[dplyr:dplyr_tidy_select]{<tidy-select>}} A selection of
columns to group by. Typically a column of speaker IDs.}

\item{.by_formant}{Whether or not the normalization method is formant
intrinsic.}

\item{.param_col}{A column identifying the DCT parameter number.}

\item{.drop_orig}{Should the originally targeted columns be dropped.}

\item{.names}{A \code{\link[glue:glue]{glue::glue()}} expression for naming the normalized
data columns. The \code{"{.formant}"} portion corresponds to the name of the original
formant columns.}

\item{.silent}{Suppress normalization information messages when running a \verb{norm_*()} function. (Defaults to \code{FALSE}, overwritable using option 'tidynorm.silent' or environment variable 'R_TIDYNORM_TIDYNORM_SILENT')}
}
\value{
A data frame of Nearey normalized DCT coefficients
}
\description{
Nearey DCT Normalization
}
\details{
\strong{Important}: This function assumes that the DCT
coefficients were estimated over log-transformed
formant values.

When formant extrinsic:
\deqn{
 \hat{F}_{ij} = \log(F_{ij}) - L
}
\deqn{
 L = \frac{1}{MN}\sum_{i=1}^M\sum_{j=1}^N \log(F_{ij})
}

When formant intrinsic:
\deqn{
 \hat{F}_{ij} = \log(F_{ij}) - L_{i}
}

\deqn{
  L_i = \frac{1}{N}\sum_{j=1}^{N}\log(F_{ij})
}

Where
\itemize{
\item \eqn{\hat{F}} is the normalized formant
\item \eqn{i} is the formant number
\item \eqn{j} is the token number
}
}
\examples{
library(tidynorm)
library(dplyr)
ggplot2_inst <- require(ggplot2)

speaker_dct <- speaker_tracks |>
  mutate(
    across(
      F1:F3,
      log
    )
  ) |>
  reframe_with_dct(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .time_col = t
  )

# Normalize DCT coefficients
speaker_dct_norm <- speaker_dct |>
  norm_dct_nearey(
    F1:F3,
    .by = speaker,
    .token_id_col = id,
    .param_col = .param
  )

# Apply average and apply inverse dct
# to plot tracks
track_norm_means <- speaker_dct_norm |>
  summarise(
    .by = c(speaker, vowel, .param),
    across(
      ends_with("_lm"),
      mean
    )
  ) |>
  reframe_with_idct(
    ends_with("_lm"),
    .by = speaker,
    .token_id_col = vowel,
    .param_col = .param
  )


if (ggplot2_inst) {
  track_norm_means |>
    ggplot(
      aes(F2_lm, F1_lm, color = speaker)
    ) +
    geom_path(
      aes(
        group = interaction(speaker, vowel)
      )
    ) +
    scale_x_reverse() +
    scale_y_reverse() +
    scale_color_brewer(palette = "Dark2") +
    coord_fixed()
}
}
\references{
Nearey, T. M. (1978). Phonetic Feature Systems for Vowels [Ph.D.].
University of Alberta.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{speaker_tracks}
\alias{speaker_tracks}
\title{Speaker Tracks}
\format{
\subsection{\code{speaker_tracks}}{

A data frame with 20,000 rows and 9 columns
\describe{
\item{speaker}{Speaker ID column}
\item{id}{Within speaker id for each token}
\item{vowel}{CMU Dictionary vowel class}
\item{plt_vclass}{Modified Labov-Trager vowel class}
\item{ipa_vclas}{IPA-like vowel class}
\item{word}{Word that the vowel appeared in}
\item{t}{Measurement time point}
\item{F1, F2, F3}{The first, second and third formants, in Hz}
}
}
}
\usage{
speaker_tracks
}
\description{
Speaker Tracks
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{options}
\alias{options}
\title{tidynorm Options}
\description{
Options to control the verbosity of tidynorm functions. The convenience function \code{\link[=tidynorm_options]{tidynorm_options()}} will set these options within your current session. For a common behavior across R sessions, set the environment variables as described below. If you've silenced informational messages and want to double check what normalization steps have been taken, use \code{\link[=check_norm]{check_norm()}}.
}
\section{Checking Option Values}{

Option values specific to \code{tidynorm} can be
accessed by passing the package name to \code{env}.

\if{html}{\out{<div class="sourceCode">}}\preformatted{options::opts(env = "tidynorm")

options::opt(x, default, env = "tidynorm")
}\if{html}{\out{</div>}}
}

\section{Options}{

\describe{
\item{tidynorm.silent}{\describe{
Suppress normalization information messages when running a \verb{norm_*()} function.\item{default: }{\preformatted{FALSE}}
\item{option: }{tidynorm.silent}
\item{envvar: }{R_TIDYNORM_TIDYNORM_SILENT (evaluated if possible, raw string otherwise)}
}}

\item{tidynorm.warnings}{\describe{
Print warnings from tidynorm functions.\item{default: }{\preformatted{TRUE}}
\item{option: }{tidynorm.warnings}
\item{envvar: }{R_TIDYNORM_TIDYNORM_WARNINGS (evaluated if possible, raw string otherwise)}
}}

}
}

\seealso{
options getOption Sys.setenv Sys.getenv
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dct.R
\name{idct}
\alias{idct}
\alias{idct.numeric}
\alias{idct.matrix}
\title{Inverse Discrete Cosine Transform}
\usage{
idct(y, n)

\method{idct}{numeric}(y, n = length(y))

\method{idct}{matrix}(y, n = nrow(y))
}
\arguments{
\item{y}{A vector or matrix of DCT coefficients}

\item{n}{The desired length of the idct}
}
\value{
The returned value depends on the values in \code{y}.

When passed a numeric vector, returns numeric vector of length \code{n}.

When passed a matrix, returns a matrix with \code{n} rows and the same number of columns as \code{y}.
}
\description{
The Inverse DCT
}
\details{
Applies the Inverse DCT (see \link{dct} for more details).

\deqn{
x_j = \sqrt{2}y_0 + 2\sum_{k=1}^{N-1} y_k \cos\left(\frac{\pi k(2j+1)}{2J}\right)
}
}
\examples{
x <- seq(0, 1, length = 10)
y <- 5 + x + (2 * (x^2)) + (-2 * (x^4))

dct_coefs <- dct(y)
recovered_y <- idct(dct_coefs)

plot(y, recovered_y)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms.R
\name{mel_to_hz}
\alias{mel_to_hz}
\title{Mel to Hz}
\usage{
mel_to_hz(mel, htk = FALSE)
}
\arguments{
\item{mel}{Numeric values in Hz}

\item{htk}{Whether or not to use the HTK formula}
}
\value{
A numeric vector of Hz values
}
\description{
Convert Mel to Hz
}
\details{
This is a direct re-implementation of the \code{hz_to_mel}
function from the \href{https://librosa.org/}{librosa} library.

The default method is to use the method due to Slaney (1998), which
is linear below 1000Hz, and logarithmic above.

If \code{htk=TRUE}, the method from HTK, due to O'Shaughnessy (1987) is used.
}
\examples{
mel_to_hz(c(7.5, 15, 25, 31))

}
\references{
McFee, B., C. Raffel, D. Liang, D. PW Ellis, M. McVicar, E. Battenberg,
and O. Nieto. librosa: Audio and music signal analysis in python.
In Proceedings of the 14th python in science conference, pp. 18-25.

O'Shaughnessy, D (1987). Speech communication: human and machine.
Addison-Wesley. p. 150. ISBN 978-0-201-16520-3.

Slaney, M. (1998) Auditory Toolbox: A MATLAB Toolbox for Auditory Modeling
Work. Technical Report, version 2, Interval Research Corporation.
}
